from openwisp_network_topology.integrations.device.apps import (
    OpenwispTopologyDeviceConfig as BaseAppConfig,
)


class OpenwispTopologyDeviceConfig(BaseAppConfig):
    name = 'openwisp2.sample_integration_device'
    label = 'sample_integration_device'



# Generated by Django 2.2.9 on 2020-01-3 13:58

from django.db import migrations

from openwisp_network_topology.migrations import migrate_addresses


class Migration(migrations.Migration):

    dependencies = [
        ('topology', '0007_create_new_address_field'),
    ]

    operations = [
        migrations.RunPython(migrate_addresses, reverse_code=migrations.RunPython.noop)
    ]



import os
import sys

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
TESTING = 'test' in sys.argv

DEBUG = True

ALLOWED_HOSTS = []

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': 'openwisp_network_topology.db',
    }
}

SECRET_KEY = '@q4z-^s=mv59#o=uutv4*m=h@)ik4%zp1)-k^_(!_7*x_&+ze$'

INSTALLED_APPS = [
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'openwisp_utils.admin_theme',
    # all-auth
    'django.contrib.sites',
    'allauth',
    'allauth.account',
    'allauth.socialaccount',
    # controller  (needed to test integration)
    'openwisp_controller.pki',
    'openwisp_controller.config',
    'reversion',
    'sortedm2m',
    'flat_json_widget',
    # network topology
    'openwisp_network_topology',
    'openwisp_network_topology.integrations.device',
    'openwisp_users',
    # admin
    'django.contrib.admin',
    # rest framework
    'rest_framework',
    'drf_yasg',
    'rest_framework.authtoken',
    'django_extensions',
    # 'debug_toolbar',
]

AUTH_USER_MODEL = 'openwisp_users.User'
SITE_ID = 1

STATICFILES_FINDERS = [
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
    'openwisp_utils.staticfiles.DependencyFinder',
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    # 'debug_toolbar.middleware.DebugToolbarMiddleware',
]

# INTERNAL_IPS = ['127.0.0.1']

ROOT_URLCONF = 'openwisp2.urls'

LANGUAGE_CODE = 'en-us'
TIME_ZONE = 'UTC'
USE_I18N = True
USE_L10N = True
USE_TZ = True
STATIC_URL = '/static/'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'OPTIONS': {
            'loaders': [
                'django.template.loaders.filesystem.Loader',
                'django.template.loaders.app_directories.Loader',
                'openwisp_utils.loaders.DependencyLoader',
            ],
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
                'openwisp_utils.admin_theme.context_processor.menu_items',
            ],
        },
    },
]

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'filters': {
        'require_debug_false': {'()': 'django.utils.log.RequireDebugFalse'},
        'require_debug_true': {'()': 'django.utils.log.RequireDebugTrue'},
    },
    'formatters': {
        'simple': {'format': '[%(levelname)s] %(message)s'},
        'verbose': {
            'format': '\n\n[%(levelname)s %(asctime)s] module: %(module)s, process: %(process)d, thread: %(thread)d\n%(message)s'
        },
    },
    'handlers': {
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'filters': ['require_debug_true'],
            'formatter': 'simple',
        },
        'mail_admins': {
            'level': 'ERROR',
            'filters': ['require_debug_false'],
            'class': 'django.utils.log.AdminEmailHandler',
        },
        'main_log': {
            'level': 'ERROR',
            'class': 'logging.handlers.RotatingFileHandler',
            'formatter': 'verbose',
            'filename': os.path.join(BASE_DIR, 'error.log'),
            'maxBytes': 5242880.0,
            'backupCount': 3,
        },
    },
    'root': {'level': 'INFO', 'handlers': ['main_log', 'console', 'mail_admins']},
    'loggers': {'py.warnings': {'handlers': ['console']}},
}

TEST_RUNNER = 'openwisp_network_topology.tests.utils.LoggingDisabledTestRunner'

EMAIL_PORT = '1025'  # for testing purposes
LOGIN_REDIRECT_URL = 'admin:index'
ACCOUNT_LOGOUT_REDIRECT_URL = LOGIN_REDIRECT_URL

OPENWISP_ORGANIZATON_USER_ADMIN = True
OPENWISP_ORGANIZATON_OWNER_ADMIN = True
OPENWISP_USERS_AUTH_API = True

CELERY_TASK_ALWAYS_EAGER = True
CELERY_TASK_EAGER_PROPAGATES = True
CELERY_BROKER_URL = 'memory://'

if not TESTING and any(['shell' in sys.argv, 'shell_plus' in sys.argv]):
    LOGGING.update(
        {
            'loggers': {
                'django.db.backends': {
                    'handlers': ['console'],
                    'level': 'DEBUG',
                    'propagate': False,
                }
            },
        }
    )

if os.environ.get('SAMPLE_APP', False):
    INSTALLED_APPS.remove('openwisp_network_topology')
    INSTALLED_APPS.remove('openwisp_network_topology.integrations.device')
    EXTENDED_APPS = ['openwisp_network_topology']
    INSTALLED_APPS += [
        'openwisp2.sample_network_topology',
        'openwisp2.sample_integration_device',
    ]
    TOPOLOGY_LINK_MODEL = 'sample_network_topology.Link'
    TOPOLOGY_NODE_MODEL = 'sample_network_topology.Node'
    TOPOLOGY_SNAPSHOT_MODEL = 'sample_network_topology.Snapshot'
    TOPOLOGY_TOPOLOGY_MODEL = 'sample_network_topology.Topology'
    TOPOLOGY_DEVICE_DEVICENODE_MODEL = 'sample_integration_device.DeviceNode'

# local settings must be imported before test runner otherwise they'll be ignored
try:
    from local_settings import *
except ImportError:
    pass



# Generated by Django 3.0.9 on 2020-09-01 14:57

import collections

import jsonfield.fields
import rest_framework.utils.encoders
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('topology', '0012_update_openvpn_netjson_ids'),
    ]

    operations = [
        migrations.AddField(
            model_name='link',
            name='user_properties',
            field=jsonfield.fields.JSONField(
                blank=True,
                default=dict,
                dump_kwargs={
                    'cls': rest_framework.utils.encoders.JSONEncoder,
                    'indent': 4,
                },
                help_text='If you need to add additional data to this link use this field',
                load_kwargs={'object_pairs_hook': collections.OrderedDict},
                verbose_name='user defined properties',
            ),
        ),
        migrations.AddField(
            model_name='node',
            name='user_properties',
            field=jsonfield.fields.JSONField(
                blank=True,
                default=dict,
                dump_kwargs={
                    'cls': rest_framework.utils.encoders.JSONEncoder,
                    'indent': 4,
                },
                help_text='If you need to add additional data to this node use this field',
                load_kwargs={'object_pairs_hook': collections.OrderedDict},
                verbose_name='user defined properties',
            ),
        ),
    ]



# Generated by Django 2.0.2 on 2018-02-19 14:50

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('topology', '0002_snapshot'),
    ]

    operations = [
        migrations.AddField(
            model_name='link',
            name='status_changed',
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AlterField(
            model_name='topology',
            name='parser',
            field=models.CharField(
                choices=[
                    ('netdiff.OlsrParser', 'OLSRd (txtinfo/jsoninfo)'),
                    ('netdiff.BatmanParser', 'batman-advanced (jsondoc/txtinfo)'),
                    ('netdiff.BmxParser', 'BMX6 (q6m)'),
                    ('netdiff.NetJsonParser', 'NetJSON NetworkGraph'),
                    ('netdiff.CnmlParser', 'CNML 1.0'),
                    ('netdiff.OpenvpnParser', 'OpenVPN'),
                ],
                help_text='Select topology format',
                max_length=128,
                verbose_name='format',
            ),
        ),
    ]



from openwisp_users.tests.test_admin import TestBasicUsersIntegration


class TestUsersIntegration(TestBasicUsersIntegration):
    """
    tests integration with openwisp_users
    """

    pass


del TestBasicUsersIntegration



import os
from io import StringIO

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group, Permission
from django.core.management import call_command
from django.test import TestCase
from swapper import load_model

User = get_user_model()
Organization = load_model('openwisp_users', 'Organization')
Topology = load_model('topology', 'Topology')


class TestUpgradeFromDjangoNetjsongraph(TestCase):
    def test_success(self):
        self.assertEqual(Topology.objects.count(), 0)
        commandOutput = StringIO()
        static_path = os.path.join(
            os.path.dirname(__file__), 'static', 'upgrader_script'
        )
        call_command(
            'upgrade_from_django_netjsongraph',
            backup=static_path,
            stdout=commandOutput,
        )
        self.assertIn('Migration Process Complete!', commandOutput.getvalue())
        os.remove(os.path.join(static_path, 'group_loaded.json'))
        os.remove(os.path.join(static_path, 'netjsongraph_loaded.json'))
        os.remove(os.path.join(static_path, 'user_loaded.json'))
        self.assertEqual(Topology.objects.count(), 1)
        try:
            Topology.objects.get(label='default')
        except Topology.DoesNotExist:
            self.fail('Topology "default" not created!')
        try:
            user = User.objects.get(username='sample')
            group = Group.objects.filter(user=user).first()
            permissions = Permission.objects.filter(user=user).first()
            self.assertEqual('sample', group.name)
            self.assertEqual('add_logentry', permissions.codename)
        except User.DoesNotExist:
            self.fail('User "sample" not created!')

    def test_files_not_found(self):
        commandOutput = StringIO()
        with self.assertRaises(FileNotFoundError):
            call_command(
                'upgrade_from_django_netjsongraph',
                backup=os.path.join(os.path.dirname(__file__)),
                stdout=commandOutput,
            )

    def test_organization_not_found(self):
        commandOutput = StringIO()
        with self.assertRaises(Organization.DoesNotExist):
            call_command(
                'upgrade_from_django_netjsongraph',
                backup=os.path.join(os.path.dirname(__file__), 'static'),
                organization='00000000-0000-0000-0000-000000000000',
                stdout=commandOutput,
            )



import swapper

Node = swapper.load_model('topology', 'Node')
Link = swapper.load_model('topology', 'Link')
Topology = swapper.load_model('topology', 'Topology')


def get_name(self):
    if hasattr(self, 'devicenode'):
        return self.devicenode.device.name
    return super(Node, self).get_name()


def topology_get_nodes_queryset(self):
    """
    Overrides Topology.get_nodes_queryset
    to avoid generating 2 additional queries for each node
    when using the name of the device as the node label
    """
    return (
        super(Topology, self)
        .get_nodes_queryset()
        .select_related('devicenode__device')
        .only(
            'id',
            'topology_id',
            'label',
            'addresses',
            'properties',
            'user_properties',
            'created',
            'modified',
            'devicenode__device__name',
        )
    )


@classmethod
def node_get_queryset(cls, qs):
    return (
        super(Node, cls)
        .get_queryset(qs)
        .select_related('devicenode__device')
        .only(
            'id',
            'topology_id',
            'topology__label',
            'topology__parser',
            'organization__id',
            'organization__name',
            'organization__is_active',
            'label',
            'addresses',
            'devicenode__device__name',
        )
    )


@classmethod
def link_get_queryset(cls, qs):
    return (
        super(Link, cls)
        .get_queryset(qs)
        .select_related('source__devicenode__device', 'target__devicenode__device')
        .only(
            'id',
            'topology_id',
            'topology__label',
            'topology__parser',
            'organization__id',
            'organization__name',
            'organization__is_active',
            'status',
            'cost',
            'cost_text',
            'source__addresses',
            'source__label',
            'source__devicenode__device__name',
            'target__addresses',
            'target__label',
            'target__devicenode__device__name',
        )
    )


Node.get_name = get_name
Node.get_queryset = node_get_queryset
Link.get_queryset = link_get_queryset
Topology.get_nodes_queryset = topology_get_nodes_queryset



import swapper

from openwisp_network_topology.integrations.device.tests import (
    TestAdmin as BaseTestAdmin,
)
from openwisp_network_topology.integrations.device.tests import (
    TestControllerIntegration as BaseTestControllerIntegration,
)
from openwisp_network_topology.integrations.device.tests import (
    TestMonitoringIntegration as BaseTestMonitoringIntegration,
)

DeviceNode = swapper.load_model('topology_device', 'DeviceNode')


class TestControllerIntegration(BaseTestControllerIntegration):
    def test_device_node_custom(self):
        self.assertTrue(DeviceNode().is_test)


class TestMonitoringIntegration(BaseTestMonitoringIntegration):
    pass


class TestAdmin(BaseTestAdmin):
    module = 'openwisp2.sample_network_topology'
    app_label = 'sample_network_topology'


del BaseTestControllerIntegration
del BaseTestMonitoringIntegration
del BaseTestAdmin



from django.db import models

from openwisp_network_topology.integrations.device.base.models import AbstractDeviceNode


class DeviceNode(AbstractDeviceNode):
    is_test = models.BooleanField(default=True)

    class Meta(AbstractDeviceNode.Meta):
        abstract = False



from unittest.mock import patch

import swapper
from django.contrib.auth.models import Permission
from django.test import TestCase
from django.urls import reverse
from rest_framework.views import APIView

from openwisp_users.tests.utils import TestOrganizationMixin

from .utils import CreateGraphObjectsMixin, CreateOrgMixin, LoadMixin, UnpublishMixin

Link = swapper.load_model('topology', 'Link')
Node = swapper.load_model('topology', 'Node')
Snapshot = swapper.load_model('topology', 'Snapshot')
Topology = swapper.load_model('topology', 'Topology')
OrganizationUser = swapper.load_model('openwisp_users', 'OrganizationUser')


class TestApi(
    CreateGraphObjectsMixin,
    CreateOrgMixin,
    UnpublishMixin,
    LoadMixin,
    TestOrganizationMixin,
    TestCase,
):
    list_url = reverse('network_collection')
    topology_model = Topology
    node_model = Node
    link_model = Link
    snapshot_model = Snapshot

    def setUp(self):
        org = self._create_org()
        t = self._create_topology(organization=org)
        user = self._create_user(username='tester', email='tester@email.com')
        perm = Permission.objects.filter(codename__endswith='topology')
        user.user_permissions.add(*perm)
        self._create_org_user(user=user, organization=org, is_admin=True)
        self._create_node(
            label="node1", addresses=["192.168.0.1"], topology=t, organization=org
        )
        self._create_node(
            label="node2", addresses=["192.168.0.2"], topology=t, organization=org
        )
        self.client.force_login(user)

    @property
    def detail_url(self):
        t = self.topology_model.objects.first()
        return reverse('network_graph', args=[t.pk])

    @property
    def receive_url(self):
        t = self.topology_model.objects.first()
        path = reverse('receive_topology', args=[t.pk])
        return f'{path}?key=test'

    @property
    def snapshot_url(self):
        t = self.topology_model.objects.first()
        path = reverse('network_graph_history', args=[t.pk])
        return f'{path}?date={self.snapshot_date}'

    def _set_receive(self):
        t = self.topology_model.objects.first()
        t.parser = 'netdiff.NetJsonParser'
        t.strategy = 'receive'
        t.key = 'test'
        t.expiration_time = 0
        t.save()

    @property
    def snapshot_date(self):
        self.topology_model.objects.first().save_snapshot()
        return self.snapshot_model.objects.first().date

    def test_list(self):
        response = self.client.get(self.list_url)
        self.assertEqual(response.data['type'], 'NetworkCollection')
        self.assertEqual(len(response.data['collection']), 1)

    def test_detail(self):
        response = self.client.get(self.detail_url)
        self.assertEqual(response.data['type'], 'NetworkGraph')

    def test_list_unpublished(self):
        self._unpublish()
        response = self.client.get(self.list_url)
        self.assertEqual(len(response.data['collection']), 0)

    def test_detail_unpublished(self):
        self._unpublish()
        response = self.client.get(self.detail_url)
        self.assertEqual(response.status_code, 404)

    def test_receive(self):
        self._set_receive()
        self.node_model.objects.all().delete()
        data = self._load('static/netjson-1-link.json')
        response = self.client.post(self.receive_url, data, content_type='text/plain')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['detail'], 'data received successfully')
        self.assertEqual(self.node_model.objects.count(), 2)
        self.assertEqual(self.link_model.objects.count(), 1)

    def test_receive_with_deprecated_url(self):
        self._set_receive()
        self.node_model.objects.all().delete()
        data = self._load('static/netjson-1-link.json')
        t = self.topology_model.objects.first()
        path = reverse('receive_topology_deprecated', args=[t.pk])
        path = f'{path}?key=test'
        response = self.client.post(path, data=data, content_type='text/plain')
        self.assertEqual(response.status_code, 200)
        expected_path = reverse('receive_topology', args=[t.pk])
        expected_path = f'{expected_path}?key=test'
        message = (
            'data received successfully. '
            'This URL is depercated and will be removed in '
            f'future versions, use {expected_path}'
        )
        self.assertEqual(response.data['detail'], message)
        self.assertEqual(self.node_model.objects.count(), 2)
        self.assertEqual(self.link_model.objects.count(), 1)

    def test_receive_404(self):
        # topology is set to FETCH strategy
        response = self.client.post(self.receive_url, content_type='text/plain')
        self.assertEqual(response.status_code, 404)

    def test_receive_415(self):
        self._set_receive()
        data = self._load('static/netjson-1-link.json')
        response = self.client.post(
            self.receive_url, data, content_type='application/xml'
        )
        self.assertEqual(response.status_code, 415)

    def test_receive_400_missing_key(self):
        self._set_receive()
        data = self._load('static/netjson-1-link.json')
        response = self.client.post(
            self.receive_url.replace('?key=test', ''), data, content_type='text/plain'
        )
        self.assertEqual(response.status_code, 400)
        self.assertIn('missing required', response.data['detail'])

    def test_receive_400_unrecognized_format(self):
        self._set_receive()
        self.node_model.objects.all().delete()
        data = 'WRONG'
        response = self.client.post(self.receive_url, data, content_type='text/plain')
        self.assertEqual(response.status_code, 400)
        self.assertIn('not recognized', response.data['detail'])

    def test_receive_403(self):
        self._set_receive()
        data = self._load('static/netjson-1-link.json')
        response = self.client.post(
            self.receive_url.replace('?key=test', '?key=wrong'),
            data,
            content_type='text/plain',
        )
        self.assertEqual(response.status_code, 403)

    def test_receive_options(self):
        self._set_receive()
        response = self.client.options(self.receive_url)
        self.assertEqual(response.data['parses'], ['text/plain'])

    def test_snapshot(self):
        response = self.client.get(self.snapshot_url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['type'], 'NetworkGraph')

    def test_snapshot_missing_date_400(self):
        date = self.snapshot_date
        response = self.client.get(
            self.snapshot_url.replace('?date={0}'.format(date), '')
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.data['detail'], 'missing required "date" parameter')

    def test_snapshot_invalid_date_403(self):
        date = self.snapshot_date
        url = self.snapshot_url.replace('?date={0}'.format(date), '?date=wrong-date')
        response = self.client.get(url)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.data['detail'], 'invalid date supplied')

    def test_snapshot_no_snapshot_404(self):
        date = self.snapshot_date
        url = self.snapshot_url.replace('?date={0}'.format(date), '?date=2001-01-01')
        response = self.client.get(url)
        self.assertEqual(response.status_code, 404)
        self.assertIn('no snapshot found', response.data['detail'])

    def _test_api_with_unauthenticated_user(self, url):
        self.client.logout()
        r = self.client.get(url)
        self.assertEqual(r.status_code, 403)
        self.assertEqual(
            r.data['detail'], 'Authentication credentials were not provided.'
        )
        self.assertEqual(len(r.data), 1)

    def _test_api_with_not_a_manager_user(self, user, url, has_detail=True):
        OrganizationUser.objects.filter(user=user).delete()
        perm = Permission.objects.filter(codename__endswith='topology')
        user.user_permissions.add(*perm)
        self.client.force_login(user)
        r = self.client.get(url)
        if not has_detail:
            self.assertEqual(r.status_code, 200)
            self.assertEqual(r.data['type'], 'NetworkCollection')
            self.assertEqual(len(r.data['collection']), 0)
            self.assertNotEqual(Topology.objects.all().count(), 0)
        else:
            detail = (
                'User is not a manager of the organization to '
                'which the requested resource belongs.'
            )
            self.assertEqual(r.status_code, 403)
            self.assertEqual(r.data['detail'], detail)
            self.assertEqual(len(r.data), 1)

    def _test_api_with_not_permitted_user(self, user, url):
        t = self.topology_model.objects.first()
        self._create_org_user(user=user, organization=t.organization, is_admin=True)
        self.client.force_login(user)
        r = self.client.get(url)
        self.assertEqual(r.status_code, 403)
        self.assertEqual(
            r.data['detail'], 'You do not have permission to perform this action.'
        )
        self.assertEqual(len(r.data), 1)

    def test_list_with_auth_enabled(self):
        user = self._create_user(username='list-user', email='list@email.com')
        with self.subTest('test api with unauthenticated user'):
            self._test_api_with_unauthenticated_user(self.list_url)

        with self.subTest('test api with not a permitted user'):
            self._test_api_with_not_permitted_user(user, self.list_url)

        with self.subTest('test api with not a member user'):
            self._test_api_with_not_a_manager_user(
                user, self.list_url, has_detail=False
            )

    def test_detail_with_auth_enabled(self):
        user = self._create_user(username='detail-user', email='detail@email.com')
        with self.subTest('test api with unauthenticated user'):
            self._test_api_with_unauthenticated_user(self.detail_url)

        with self.subTest('test api with not a permitted user'):
            self._test_api_with_not_permitted_user(user, self.detail_url)

        with self.subTest('test api with not a member user'):
            self._test_api_with_not_a_manager_user(user, self.detail_url)

    def test_snapshot_with_auth_enabled(self):
        user = self._create_user(username='snapshot-user', email='snapshot@email.com')
        with self.subTest('test api with unauthenticated user'):
            self._test_api_with_unauthenticated_user(self.snapshot_url)

        with self.subTest('test api with not a permitted user'):
            self._test_api_with_not_permitted_user(user, self.snapshot_url)

        with self.subTest('test api with not a member user'):
            self._test_api_with_not_a_manager_user(user, self.snapshot_url)

    def _successful_api_tests(self):
        with self.subTest('test list'):
            r = self.client.get(self.list_url)
            self.assertEqual(r.status_code, 200)
            self.assertEqual(r.data['type'], 'NetworkCollection')
            self.assertEqual(len(r.data['collection']), 1)

        with self.subTest('test detail'):
            response = self.client.get(self.detail_url)
            self.assertEqual(response.data['type'], 'NetworkGraph')

        with self.subTest('test receive'):
            self._set_receive()
            self.node_model.objects.all().delete()
            data = self._load('static/netjson-1-link.json')
            response = self.client.post(
                self.receive_url, data, content_type='text/plain'
            )
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.data['detail'], 'data received successfully')
            self.assertEqual(self.node_model.objects.count(), 2)
            self.assertEqual(self.link_model.objects.count(), 1)

        with self.subTest('test history'):
            response = self.client.get(self.snapshot_url)
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.data['type'], 'NetworkGraph')

    @patch.object(APIView, 'get_permissions', return_value=[])
    @patch.object(APIView, 'get_authenticators', return_value=[])
    def test_api_with_auth_disabled(self, perm_mocked, auth_mocked):
        user = self._get_user(username='tester')
        self.client.logout()
        self._successful_api_tests()
        self.client.force_login(user)

    def test_superuser_with_api_auth_enabled(self):
        user = self._create_admin(username='superapi', email='superapi@email.com')
        self.client.force_login(user)
        self._successful_api_tests()

    @patch.object(APIView, 'get_permissions', return_value=[])
    @patch.object(APIView, 'get_authenticators', return_value=[])
    def test_superuser_with_api_auth_disabled(self, perm_mocked, auth_mocked):
        user = self._create_admin(username='superapi', email='superapi@email.com')
        self.client.force_login(user)
        self._successful_api_tests()



import json
from collections import OrderedDict
from copy import deepcopy
from datetime import timedelta

import swapper
from django.db import models
from django.utils.functional import cached_property
from django.utils.timezone import now
from django.utils.translation import ugettext_lazy as _
from jsonfield import JSONField
from rest_framework.utils.encoders import JSONEncoder

from openwisp_users.mixins import OrgMixin
from openwisp_utils.base import TimeStampedEditableModel

from .. import settings as app_settings
from ..utils import print_info


class AbstractNode(OrgMixin, TimeStampedEditableModel):
    """
    NetJSON NetworkGraph Node Object implementation
    """

    topology = models.ForeignKey(
        swapper.get_model_name('topology', 'Topology'), on_delete=models.CASCADE
    )
    label = models.CharField(max_length=64, blank=True)
    # netjson ID and local_addresses
    addresses = JSONField(default=[], blank=True)
    properties = JSONField(
        default=dict,
        blank=True,
        load_kwargs={'object_pairs_hook': OrderedDict},
        dump_kwargs={'indent': 4, 'cls': JSONEncoder},
    )
    user_properties = JSONField(
        verbose_name=_('user defined properties'),
        help_text=_('If you need to add additional data to this node use this field'),
        default=dict,
        blank=True,
        load_kwargs={'object_pairs_hook': OrderedDict},
        dump_kwargs={'indent': 4, 'cls': JSONEncoder},
    )

    class Meta:
        abstract = True

    def __str__(self):
        return self.name

    def full_clean(self, *args, **kwargs):
        self.organization = self.topology.organization
        return super().full_clean(*args, **kwargs)

    def clean(self):

        if self.properties is None:
            self.properties = {}

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)

    @property
    def netjson_id(self):
        if len(self.addresses) > 0:
            return self.addresses[0]

    @cached_property
    def local_addresses(self):
        if len(self.addresses) > 1:
            return self.addresses[1:]

    @property
    def name(self):
        return self.label or self.netjson_id or ''

    def get_name(self):
        """
        May be overridden/monkey patched to get the node name
        from other sources (e.g: device name in openwisp-controller)
        """
        return self.name

    def json(self, dict=False, original=False, **kwargs):
        """
        Returns a NetJSON NetworkGraph Node object.

        If ``original`` is passed, the data will be returned
        as it has been collected from the network (used when
        doing the comparison).
        """
        netjson = OrderedDict({'id': self.netjson_id})
        label = self.get_name()
        if label:
            netjson['label'] = label
        for attr in ['local_addresses', 'properties']:
            value = getattr(self, attr)
            if value or attr == 'properties':
                netjson[attr] = deepcopy(value)
        if not original:
            netjson['properties'].update(deepcopy(self.user_properties))
            netjson['properties']['created'] = JSONEncoder().default(self.created)
            netjson['properties']['modified'] = JSONEncoder().default(self.modified)
        if dict:
            return netjson
        return json.dumps(netjson, cls=JSONEncoder, **kwargs)

    @classmethod
    def get_from_address(cls, address, topology):
        """
        Find node from one of its addresses and its topology.
        :param address: string
        :param topology: Topology instance
        :returns: Node object or None
        """
        address = '"{}"'.format(address)
        return cls.objects.filter(
            topology=topology, addresses__contains=address
        ).first()

    @classmethod
    def count_address(cls, address, topology):
        """
        Count nodes with the specified address and topology.
        :param address: string
        :param topology: Topology instance
        :returns: int
        """
        address = '"{}"'.format(address)
        return cls.objects.filter(
            topology=topology, addresses__contains=address
        ).count()

    @classmethod
    def delete_expired_nodes(cls):
        """
        deletes nodes that have not been  connected to the network
        for more than ``NETJSONGRAPH__EXPIRATION`` days
        """
        NODE_EXPIRATION = app_settings.NODE_EXPIRATION
        LINK_EXPIRATION = app_settings.LINK_EXPIRATION
        if NODE_EXPIRATION not in [False, None] and LINK_EXPIRATION not in [
            False,
            None,
        ]:
            expiration_date = now() - timedelta(days=int(NODE_EXPIRATION))
            expired_nodes = cls.objects.filter(
                modified__lt=expiration_date,
                source_link_set__isnull=True,
                target_link_set__isnull=True,
            )
            expired_nodes_length = len(expired_nodes)
            if expired_nodes_length:
                print_info('Deleting {0} expired nodes'.format(expired_nodes_length))
                for node in expired_nodes:
                    node.delete()

    @classmethod
    def get_queryset(cls, qs):
        """ admin list queryset """
        return qs.select_related('organization', 'topology')



from openwisp_network_topology.api.views import (
    NetworkCollectionView as BaseNetworkCollectionView,
)
from openwisp_network_topology.api.views import (
    NetworkGraphHistoryView as BaseNetworkGraphHistoryView,
)
from openwisp_network_topology.api.views import NetworkGraphView as BaseNetworkGraphView
from openwisp_network_topology.api.views import (
    ReceiveTopologyView as BaseReceiveTopologyView,
)


class NetworkCollectionView(BaseNetworkCollectionView):
    pass


class NetworkGraphView(BaseNetworkGraphView):
    pass


class ReceiveTopologyView(BaseReceiveTopologyView):
    pass


class NetworkGraphHistoryView(BaseNetworkGraphHistoryView):
    pass


network_collection = NetworkCollectionView.as_view()
network_graph = NetworkGraphView.as_view()
network_graph_history = NetworkGraphHistoryView.as_view()
receive_topology = ReceiveTopologyView.as_view()



import swapper
from django.core.management.base import BaseCommand

Topology = swapper.load_model('topology', 'Topology')


class BaseUpdateCommand(BaseCommand):
    help = 'Update network topology'
    topology_model = Topology

    def add_arguments(self, parser):
        parser.add_argument(
            '--label',
            action='store',
            default=None,
            help='Will update topologies containing label',
        )

    def handle(self, *args, **options):
        self.topology_model.update_all(options['label'])


class BaseSaveSnapshotCommand(BaseCommand):
    help = 'Save network topology graph snapshot'
    topology_model = Topology

    def add_arguments(self, parser):
        parser.add_argument(
            '--label',
            action='store',
            default=None,
            help='Will save snapshots of topologies containing label',
        )

    def handle(self, *args, **options):
        self.topology_model.save_snapshot_all(options['label'])



from collections import OrderedDict

import swapper
from rest_framework import serializers


class NetworkCollectionSerializer(serializers.ListSerializer):
    """
    NetJSON NetworkCollection
    """

    @property
    def data(self):
        return super(serializers.ListSerializer, self).data

    def to_representation(self, data):
        return OrderedDict(
            (
                ('type', 'NetworkCollection'),
                ('collection', super().to_representation(data)),
            )
        )


class NetworkGraphSerializer(serializers.ModelSerializer):
    """
    NetJSON NetworkGraph
    """

    def to_representation(self, obj):
        return obj.json(dict=True)

    class Meta:
        model = swapper.load_model('topology', 'Topology')
        fields = '__all__'
        list_serializer_class = NetworkCollectionSerializer



from importlib import import_module

import swapper
from django.apps import AppConfig
from django.db import transaction
from django.db.models.signals import post_save
from django.utils.translation import ugettext_lazy as _

from ...utils import link_status_changed
from .tasks import create_device_node_relation, trigger_device_updates


class OpenwispTopologyDeviceConfig(AppConfig):
    name = 'openwisp_network_topology.integrations.device'
    label = 'topology_device'
    verbose_name = _('Topology Device Integration')

    def ready(self):
        self.connect_signals()
        self.override_node_label()

    def connect_signals(self):
        Node = swapper.load_model('topology', 'Node')
        Link = swapper.load_model('topology', 'Link')

        post_save.connect(
            self.create_device_rel, sender=Node, dispatch_uid='node_to_device_rel',
        )
        link_status_changed.connect(
            self.link_status_changed_receiver,
            sender=Link,
            dispatch_uid='controller_integration_link_status_chaged',
        )

    @classmethod
    def create_device_rel(cls, instance, created, **kwargs):
        if not created:
            return
        transaction.on_commit(lambda: create_device_node_relation.delay(instance.pk))

    @classmethod
    def link_status_changed_receiver(cls, link, **kwargs):
        transaction.on_commit(lambda: trigger_device_updates.delay(link.pk))

    def override_node_label(self):
        import_module('openwisp_network_topology.integrations.device.overrides')



import logging

from django.conf import settings

# 'pre_django_setup' is supposed to be a logger
# that can work before registered Apps are
# ready in django.setup() process.
logger = logging.getLogger('pre_django_setup')


def get_settings_value(option, default):
    outdated_option_used = False
    if hasattr(settings, f'NETJSONGRAPH_{option}'):
        outdated_option_used = 'NETJSONGRAPH'
    elif hasattr(settings, f'TOPOLOGY_{option}'):
        outdated_option_used = 'TOPOLOGY'
    if outdated_option_used:
        logger.warn(
            f'{outdated_option_used}_{option} setting is deprecated. It will be removed '
            f'in the future, please use OPENWISP_NETWORK_TOPOLOGY_{option} instead.'
        )
        return getattr(settings, f'{outdated_option_used}_{option}')
    return getattr(settings, f'OPENWISP_NETWORK_TOPOLOGY_{option}', default)


DEFAULT_PARSERS = [
    ('netdiff.OlsrParser', 'OLSRd (txtinfo/jsoninfo)'),
    ('netdiff.BatmanParser', 'batman-advanced (jsondoc/txtinfo)'),
    ('netdiff.BmxParser', 'BMX6 (q6m)'),
    ('netdiff.NetJsonParser', 'NetJSON NetworkGraph'),
    ('netdiff.CnmlParser', 'CNML 1.0'),
    ('netdiff.OpenvpnParser', 'OpenVPN'),
]

PARSERS = DEFAULT_PARSERS + get_settings_value('PARSERS', [])
SIGNALS = get_settings_value('SIGNALS', None)
TIMEOUT = get_settings_value('TIMEOUT', 8)
LINK_EXPIRATION = get_settings_value('LINK_EXPIRATION', 60)
NODE_EXPIRATION = get_settings_value('NODE_EXPIRATION', False)
VISUALIZER_CSS = get_settings_value('VISUALIZER_CSS', 'netjsongraph/css/style.css')
TOPOLOGY_API_URLCONF = get_settings_value('API_URLCONF', None)
TOPOLOGY_API_BASEURL = get_settings_value('API_BASEURL', None)
TOPOLOGY_API_AUTH_REQUIRED = get_settings_value('API_AUTH_REQUIRED', True)



import json
import logging

import swapper
from django.core.exceptions import ValidationError
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _
from netdiff.exceptions import NetdiffException
from rest_framework import generics
from rest_framework.authentication import BasicAuthentication, SessionAuthentication
from rest_framework.permissions import (
    BasePermission,
    DjangoModelPermissions,
    IsAuthenticated,
)
from rest_framework.response import Response
from rest_framework.views import APIView

from openwisp_users.api.authentication import BearerAuthentication
from openwisp_users.api.permissions import IsOrganizationManager

from .. import settings as app_settings
from ..utils import get_object_or_404
from .parsers import TextParser
from .serializers import NetworkGraphSerializer

logger = logging.getLogger(__name__)
Snapshot = swapper.load_model('topology', 'Snapshot')
Topology = swapper.load_model('topology', 'Topology')


class HasGetMethodPermission(BasePermission):
    def has_permission(self, request, view):
        return self.check_permission(request)

    def has_object_permission(self, request, view, obj):
        return self.check_permission(request)

    def check_permission(self, request):
        user = request.user
        app_label = Topology._meta.app_label.lower()
        model = Topology._meta.object_name.lower()
        change_perm = f'{app_label}.change_{model}'
        view_perm = f'{app_label}.view_{model}'
        if user.is_authenticated:
            if user.is_superuser or request.method != 'GET':
                return True
            if request.method == 'GET' and (
                user.has_permission(change_perm) or user.has_permission(view_perm)
            ):
                return True
        return False


class RequireAuthentication(APIView):

    if app_settings.TOPOLOGY_API_AUTH_REQUIRED:
        authentication_classes = [
            SessionAuthentication,
            BasicAuthentication,
            BearerAuthentication,
        ]
        permission_classes = [
            IsAuthenticated,
            IsOrganizationManager,
            DjangoModelPermissions,
            HasGetMethodPermission,
        ]


class NetworkCollectionView(generics.ListAPIView, RequireAuthentication):
    """
    Data of all the topologies returned
    in NetJSON NetworkCollection format
    """

    serializer_class = NetworkGraphSerializer
    queryset = Topology.objects.filter(published=True)

    def list(self, request, *args, **kwargs):
        self.check_permissions(request)
        return super().list(request, *args, **kwargs)

    def get_queryset(self):
        queryset = super().get_queryset()
        user = self.request.user
        if user.is_superuser:
            return queryset
        if user.is_authenticated:
            queryset = queryset.filter(organization__in=user.organizations_managed)
        return queryset


class NetworkGraphView(generics.RetrieveAPIView, RequireAuthentication):
    """
    Data of a specific topology returned
    in NetJSON NetworkGraph format
    """

    serializer_class = NetworkGraphSerializer
    queryset = Topology.objects.filter(published=True)


class ReceiveTopologyView(APIView):
    """
    This views allow nodes to send topology data using the RECEIVE strategy.

    Required query string parameters:
        * key

    Allowed content-types:
        * text/plain
    """

    model = Topology
    parser_classes = (TextParser,)

    def post(self, request, pk, format=None):
        topology = get_object_or_404(self.model, pk, strategy='receive')
        key = request.query_params.get('key')
        # wrong content type: 415
        if request.content_type != 'text/plain':
            return Response(
                {'detail': _('expected content type "text/plain"')}, status=415
            )
        # missing key: 400
        if not key:
            return Response(
                {'detail': _('missing required "key" parameter')}, status=400
            )
        # wrong key 403
        if topology.key != key:
            return Response({'detail': _('wrong key')}, status=403)
        try:
            topology.receive(request.data)
        except NetdiffException as e:
            error = _(
                'Supplied data not recognized as %s, '
                'got exception of type "%s" '
                'with message "%s"'
            ) % (topology.get_parser_display(), e.__class__.__name__, e)
            return Response({'detail': error}, status=400)
        success_message = _('data received successfully')
        deprecated_url = reverse('receive_topology_deprecated', args=[pk])
        if request.path == deprecated_url:
            expected_path = reverse('receive_topology', args=[pk])
            expected_path = f'{expected_path}?key={key}'
            warning = _(
                'This URL is depercated and will be removed in '
                f'future versions, use {expected_path}'
            )
            logger.warning(warning)
            success_message = f'{success_message}. {warning}'
        return Response({'detail': success_message})


class NetworkGraphHistoryView(RequireAuthentication):
    """
    History of a specific topology returned
    in NetJSON NetworkGraph format
    """

    topology_model = Topology
    snapshot_model = Snapshot
    queryset = topology_model.objects.all()  # Required for DjangoModelPermissions

    def get(self, request, pk, format=None):
        topology = get_object_or_404(self.topology_model, pk)
        self.check_object_permissions(request, topology)
        date = request.query_params.get('date')
        options = dict(topology=topology, date=date)
        # missing date: 400
        if not date:
            return Response(
                {'detail': _('missing required "date" parameter')}, status=400
            )
        try:
            s = self.snapshot_model.objects.get(**options)
            return Response(json.loads(s.data))
        except self.snapshot_model.DoesNotExist:
            return Response(
                {'detail': _('no snapshot found for this date')}, status=404
            )
        except ValidationError:
            return Response({'detail': _('invalid date supplied')}, status=403)


network_collection = NetworkCollectionView.as_view()
network_graph = NetworkGraphView.as_view()
network_graph_history = NetworkGraphHistoryView.as_view()
receive_topology = ReceiveTopologyView.as_view()



import logging
from ipaddress import ip_address

from django.conf import settings
from django.db import models
from django.utils.module_loading import import_string
from swapper import get_model_name, load_model

from openwisp_utils.base import UUIDModel

logger = logging.getLogger(__name__)

trigger_device_checks_path = 'openwisp_monitoring.device.tasks.trigger_device_checks'


class AbstractDeviceNode(UUIDModel):
    node = models.OneToOneField(
        get_model_name('topology', 'Node'), on_delete=models.CASCADE,
    )
    device = models.ForeignKey(
        get_model_name('config', 'Device'), on_delete=models.CASCADE,
    )

    # relations will be auto-created only for these parsers
    ENABLED_PARSERS = {
        'netdiff.OpenvpnParser': {
            'auto_create': 'auto_create_openvpn',
            'link_down': 'link_down_openvpn',
            'link_up': 'link_up_openvpn',
        }
    }

    class Meta:
        abstract = True
        unique_together = ('node', 'device')

    @classmethod
    def auto_create(cls, node):
        """
        Attempts to perform automatic creation of DeviceNode objects.
        The right action to perform depends on the Topology used.
        """
        opts = cls.ENABLED_PARSERS.get(node.topology.parser)
        if opts:
            return getattr(cls, opts['auto_create'])(node)

    @classmethod
    def auto_create_openvpn(cls, node):
        """
        Implementation of the integration between
        controller and network-topology modules
        when using OpenVPN (using the common name)
        """
        common_name = node.properties.get('common_name')
        if not common_name:
            return

        Device = load_model('config', 'Device')
        device = (
            Device.objects.only('id', 'name')
            .filter(
                organization_id=node.organization_id,
                config__vpnclient__cert__common_name=common_name,
            )
            .first()
        )
        if not device:
            return

        device_node = cls(device=device, node=node)
        try:
            device_node.full_clean()
            device_node.save()
        except Exception:
            logger.exception('Exception raised during auto_create_openvpn')
            return
        else:
            logger.info(f'DeviceNode relation created for {node.label} - {device.name}')
            return device_node

    def link_action(self, link, status):
        """
        Performs clean-up operations when link goes down.
        The right action to perform depends on the Topology used.
        """
        opts = self.ENABLED_PARSERS.get(link.topology.parser)
        if opts:
            key = f'link_{status}'
            return getattr(self, opts[key])()

    def link_down_openvpn(self):
        """
        Link down action for OpenVPN
        """
        self.device.management_ip = None
        self.device.save()

    def link_up_openvpn(self):
        """
        Link up action for OpenVPN
        """
        addresses = self.node.addresses
        try:
            address = ip_address(addresses[1])
        except (KeyError, ValueError) as e:
            addresses = ', '.join(addresses)
            logger.warning(
                f'{e.__class__.__name__} raised while processing addresses: {addresses}'
            )
        else:
            self.device.management_ip = str(address)
            self.device.save()

    @classmethod
    def filter_by_link(cls, link):
        """
        Returns a queryset which looks for a DeviceNode which is related
        to the specified Link instance.
        """
        return cls.objects.filter(
            models.Q(node__source_link_set__pk=link.pk)
            | models.Q(node__target_link_set__pk=link.pk)
        ).select_related('device', 'node')

    @classmethod
    def trigger_device_updates(cls, link):
        """
        Used to refresh controller and monitoring information
        whenever the status of a link changes
        """
        if link.topology.parser not in cls.ENABLED_PARSERS:
            return
        for device_node in cls.filter_by_link(link):
            device_node.link_action(link, link.status)
            # triggers monitoring checks if OpenWISP Monitoring is enabled
            if 'openwisp_monitoring.device' in settings.INSTALLED_APPS:
                run_checks = import_string(trigger_device_checks_path)
                run_checks.delay(device_node.device.pk, recovery=link.status == 'up')



import swapper
from django.core.exceptions import ValidationError
from django.test import TestCase

from openwisp_utils.tests import catch_signal

from ..utils import link_status_changed
from .utils import CreateGraphObjectsMixin, CreateOrgMixin

Link = swapper.load_model('topology', 'Link')
Node = swapper.load_model('topology', 'Node')
Topology = swapper.load_model('topology', 'Topology')


class TestLink(CreateOrgMixin, CreateGraphObjectsMixin, TestCase):
    topology_model = Topology
    node_model = Node
    link_model = Link

    def setUp(self):
        org = self._create_org()
        t = self._create_topology(organization=org)
        self._create_node(
            label='node1', addresses=['192.168.0.1'], topology=t, organization=org
        )
        self._create_node(
            label='node2', addresses=['192.168.0.2'], topology=t, organization=org
        )

    def _get_nodes(self):
        return self.node_model.objects.all()

    def test_str(self):
        t = self.topology_model.objects.first()
        node1, node2 = self._get_nodes()
        link = t._create_link(source=node1, target=node2, cost=1.0)
        self.assertIsInstance(str(link), str)

    def test_clean_properties(self):
        t = self.topology_model.objects.first()
        node1, node2 = self._get_nodes()
        link = t._create_link(source=node1, target=node2, cost=1.0, properties=None)
        link.full_clean()
        self.assertEqual(link.properties, {})

    def test_clean_properties_status(self):
        t = self.topology_model.objects.first()
        node1, node2 = self._get_nodes()
        link = t._create_link(
            source=node1, target=node2, cost=1.0, properties={'status': 'up'}
        )
        link.full_clean()
        # status must not be saved in properties or it will override the real status
        self.assertEqual(link.properties, {})

    def test_same_source_and_target_id(self):
        t = self.topology_model.objects.first()
        node_id = self.node_model.objects.first().pk
        link = t._create_link(source_id=node_id, target_id=node_id, cost=1)
        with self.assertRaises(ValidationError):
            link.full_clean()

    def test_same_source_and_target(self):
        t = self.topology_model.objects.first()
        node = self.node_model.objects.first()
        link = t._create_link(source=node, target=node, cost=1)
        with self.assertRaises(ValidationError):
            link.full_clean()

    def test_json(self):
        t = self.topology_model.objects.first()
        node1, node2 = self._get_nodes()
        link = t._create_link(
            source=node1,
            target=node2,
            cost=1.0,
            cost_text='100mbit/s',
            properties={"pretty": True},
        )
        self.assertEqual(
            dict(link.json(dict=True)),
            {
                'source': '192.168.0.1',
                'target': '192.168.0.2',
                'cost': 1.0,
                'cost_text': '100mbit/s',
                'properties': {
                    'pretty': True,
                    'status': 'up',
                    'created': link.created,
                    'modified': link.modified,
                    'status_changed': link.status_changed,
                },
            },
        )
        self.assertIsInstance(link.json(), str)
        link.status = 'down'
        link.save()
        link.refresh_from_db()
        netjson = link.json(dict=True)
        with self.subTest('status should not be saved in properties'):
            self.assertNotIn('status', link.properties)
        with self.subTest('status should not be overridden'):
            self.assertEqual(netjson['properties']['status'], 'down')
        with self.subTest('testing original=True'):
            netjson = link.json(dict=True, original=True)
            self.assertNotIn('status', netjson['properties'])
            self.assertNotIn('status_changed', netjson['properties'])
            self.assertNotIn('created', netjson['properties'])
            self.assertNotIn('modified', netjson['properties'])

    def test_get_from_nodes(self):
        t = self.topology_model.objects.first()
        node1, node2 = self._get_nodes()
        link = t._create_link(
            source=node1,
            target=node2,
            cost=1.0,
            cost_text='100mbit/s',
            properties='{"pretty": true}',
        )
        link.full_clean()
        link.save()
        link = self.link_model.get_from_nodes('192.168.0.1', '192.168.0.2', t)
        self.assertIsInstance(link, self.link_model)
        link = self.link_model.get_from_nodes('wrong', 'wrong', t)
        self.assertIsNone(link)

    def test_status_change_signal_sent(self):
        self.signal_was_called = False
        t = self.topology_model.objects.first()
        node1, node2 = self._get_nodes()
        with catch_signal(link_status_changed) as handler:
            link = t._create_link(source=node1, target=node2, cost=1.0, status='up')
            link.save()
            handler.assert_not_called()
        with catch_signal(link_status_changed) as handler:
            link.status = 'down'
            link.save()
        handler.assert_called_once_with(
            link=link, sender=self.link_model, signal=link_status_changed,
        )
        self.assertEqual(link.status, 'down')

    def test_link_auto_org(self):
        t = self.topology_model.objects.first()
        node1, node2 = self._get_nodes()
        link = self.link_model(
            source=node1, target=node2, cost=1.0, cost_text='100mbit/s', topology=t
        )
        link.full_clean()
        self.assertEqual(link.organization, t.organization)

    def test_user_properties_in_json(self):
        t = self.topology_model.objects.first()
        node1, node2 = self._get_nodes()
        link = self.link_model(
            source=node1, target=node2, cost=1.0, cost_text='100mbit/s', topology=t
        )
        link.properties = {
            'wired': True,
        }
        link.user_properties = {
            'user_property': True,
        }
        link.full_clean()
        link.save()

        with self.subTest('view json when original is False'):
            data = link.json(dict=True)
            self.assertIn('wired', data['properties'])
            self.assertIn('user_property', data['properties'])

        with self.subTest('view json when original is True'):
            data = link.json(dict=True, original=True)
            self.assertIn('wired', data['properties'])
            self.assertNotIn('user_property', data['properties'])



from django.urls import reverse

from ..settings import TOPOLOGY_API_BASEURL, TOPOLOGY_API_URLCONF


class GraphVisualizerUrls:
    def get_graph_urls(self, request, pk):
        graph_path = reverse('network_graph', urlconf=TOPOLOGY_API_URLCONF, args=[pk])
        history_path = reverse(
            'network_graph_history', urlconf=TOPOLOGY_API_URLCONF, args=[pk]
        )
        if TOPOLOGY_API_BASEURL:
            graph_url = '{}{}'.format(TOPOLOGY_API_BASEURL, graph_path)
            history_url = '{}{}'.format(TOPOLOGY_API_BASEURL, history_path)
        else:
            graph_url = '{0}://{1}{2}'.format(
                request.scheme, request.get_host(), graph_path
            )
            history_url = '{0}://{1}{2}'.format(
                request.scheme, request.get_host(), history_path
            )
        return graph_url, history_url



import swapper
from django import forms
from django.conf.urls import url
from django.contrib import admin, messages
from django.contrib.admin import ModelAdmin
from django.db.models import Q
from django.template.response import TemplateResponse
from django.templatetags.static import static
from django.urls import reverse
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _
from flat_json_widget.widgets import FlatJsonWidget

from openwisp_users.multitenancy import (
    MultitenantAdminMixin,
    MultitenantOrgFilter,
    MultitenantRelatedOrgFilter,
)
from openwisp_utils.admin import ReceiveUrlAdmin, UUIDAdmin

from . import settings as app_settings
from .contextmanagers import log_failure
from .visualizer import GraphVisualizerUrls

Link = swapper.load_model('topology', 'Link')
Node = swapper.load_model('topology', 'Node')
Topology = swapper.load_model('topology', 'Topology')


class TimeStampedEditableAdmin(ModelAdmin):
    """
    ModelAdmin for TimeStampedEditableModel
    """

    def __init__(self, *args, **kwargs):
        self.readonly_fields += (
            'created',
            'modified',
        )
        super().__init__(*args, **kwargs)


class BaseAdmin(TimeStampedEditableAdmin):
    save_on_top = True

    class Media:
        css = {
            'all': [
                static('netjsongraph/css/src/netjsongraph.css'),
                static('netjsongraph/css/lib/jquery-ui.min.css'),
                static('netjsongraph/css/style.css'),
                static('netjsongraph/css/admin.css'),
            ]
        }
        js = [
            static('netjsongraph/js/lib/d3.min.js'),
            static('netjsongraph/js/lib/jquery-ui.min.js'),
            static('netjsongraph/js/src/netjsongraph.js'),
            static('netjsongraph/js/strategy-switcher.js'),
            static('netjsongraph/js/topology-history.js'),
            static('netjsongraph/js/visualize.js'),
        ]


@admin.register(Topology)
class TopologyAdmin(
    MultitenantAdminMixin, GraphVisualizerUrls, UUIDAdmin, ReceiveUrlAdmin, BaseAdmin
):
    model = Topology
    list_display = [
        'label',
        'parser',
        'strategy',
        'published',
        'created',
        'modified',
        'organization',
    ]
    readonly_fields = [
        'uuid',
        'protocol',
        'version',
        'revision',
        'metric',
        'receive_url',
    ]
    list_filter = [
        'parser',
        'strategy',
        ('organization', MultitenantOrgFilter),
    ]
    search_fields = ['label', 'id']
    actions = ['update_selected', 'unpublish_selected', 'publish_selected']
    fields = [
        'uuid',
        'label',
        'organization',
        'parser',
        'strategy',
        'url',
        'key',
        'expiration_time',
        'receive_url',
        'published',
        'protocol',
        'version',
        'revision',
        'metric',
        'created',
    ]
    receive_url_name = 'receive_topology'
    receive_url_urlconf = app_settings.TOPOLOGY_API_URLCONF
    receive_url_baseurl = app_settings.TOPOLOGY_API_BASEURL

    def get_actions(self, request):
        """
        move delete action to last position
        """
        actions = super().get_actions(request)
        if 'delete_selected' in actions:
            delete = actions['delete_selected']
            del actions['delete_selected']
            actions['delete_selected'] = delete
        return actions

    def change_view(self, request, object_id, form_url='', extra_context=None):
        extra_context = extra_context or {}
        prefix = 'admin:{0}_{1}'.format(
            self.opts.app_label, self.model.__name__.lower()
        )
        text = _('View topology graph')
        extra_context.update(
            {
                'additional_buttons': [
                    {
                        'type': 'button',
                        'url': reverse(
                            '{0}_visualize'.format(prefix), args=[object_id]
                        ),
                        'class': 'visualizelink',
                        'value': text,
                        'title': '{0} (ALT+P)'.format(text),
                    }
                ]
            }
        )
        return super().change_view(request, object_id, form_url, extra_context)

    def get_urls(self):
        options = getattr(self.model, '_meta')
        url_prefix = '{0}_{1}'.format(options.app_label, options.model_name)
        return [
            url(
                r'^visualize/(?P<pk>[^/]+)/$',
                self.admin_site.admin_view(self.visualize_view),
                name='{0}_visualize'.format(url_prefix),
            ),
        ] + super().get_urls()

    def _message(self, request, rows, suffix, level=messages.SUCCESS):
        if rows == 1:
            prefix = _('1 {0} was'.format(self.model._meta.verbose_name))
        else:  # pragma: nocover
            prefix = _(
                '{0} {1} were'.format(rows, self.model._meta.verbose_name_plural)
            )
        self.message_user(request, '{0} {1}'.format(prefix, suffix), level=level)

    def update_selected(self, request, queryset):
        items = list(queryset)
        failed = []
        ignored = []
        for item in items:
            if item.strategy == 'fetch':
                try:
                    item.update()
                except Exception as e:
                    failed.append('{0}: {1}'.format(item.label, str(e)))
                    with log_failure('update topology admin action', item):
                        raise e
            else:
                ignored.append(item)
        # remove item from items if ignored.
        for item in ignored:
            if item in items:
                items.remove(item)
        failures = len(failed)
        successes = len(items) - failures
        total_ignored = len(ignored)
        if successes > 0:
            self._message(request, successes, _('successfully updated'))
        if failures > 0:
            message = _('not updated. %s') % '; '.join(failed)
            self._message(request, failures, message, level=messages.ERROR)
        if total_ignored > 0:
            message = _("ignored (not using FETCH strategy)")
            self._message(request, total_ignored, message, level=messages.WARNING)

    update_selected.short_description = _(
        'Update selected topologies (FETCH strategy only)'
    )

    def publish_selected(self, request, queryset):
        rows_updated = queryset.update(published=True)
        self._message(request, rows_updated, _('successfully published'))

    publish_selected.short_description = _('Publish selected topologies')

    def unpublish_selected(self, request, queryset):
        rows_updated = queryset.update(published=False)
        self._message(request, rows_updated, _('successfully unpublished'))

    unpublish_selected.short_description = _('Unpublish selected items')

    def visualize_view(self, request, pk):
        graph_url, history_url = self.get_graph_urls(request, pk)
        context = self.admin_site.each_context(request)
        opts = self.model._meta
        context.update(
            {
                'is_popup': True,
                'opts': opts,
                'change': False,
                'media': self.media,
                'graph_url': graph_url,
                'history_url': history_url,
            }
        )
        return TemplateResponse(request, 'admin/topology/visualize.html', context)


class UserPropertiesForm(forms.ModelForm):
    class Meta:
        widgets = {
            'user_properties': FlatJsonWidget,
        }


class NodeLinkMixin(MultitenantAdminMixin):
    """
    Hides organization in add form
    Shows organization readonly in change form
    Adds get_queryset class method that enables
    other extensions to override the admin queryset
    """

    readonly_fields = ['organization', 'readonly_properties']

    def get_fields(self, request, obj):
        fields = super().get_fields(request, obj)
        if obj:
            return fields
        fields_copy = list(fields)
        fields_copy.remove('organization')
        return fields_copy

    def get_queryset(self, request):
        return self.model.get_queryset(super().get_queryset(request))

    def readonly_properties(self, obj):
        output = ''
        for key, value in obj.properties.items():
            key = key.replace('_', ' ').capitalize()
            output += f'<p><strong>{key}</strong>: {value}</p>'
        return mark_safe(output)

    readonly_properties.short_description = _('Properties')


@admin.register(Node)
class NodeAdmin(NodeLinkMixin, BaseAdmin):
    model = Node
    form = UserPropertiesForm
    change_form_template = 'admin/topology/node/change_form.html'
    list_display = ['get_name', 'organization', 'topology', 'addresses']
    search_fields = ['addresses', 'label', 'properties']
    list_filter = [
        ('organization', MultitenantOrgFilter),
        ('topology', MultitenantRelatedOrgFilter),
    ]
    multitenant_shared_relations = ('topology',)
    fields = [
        'topology',
        'organization',
        'label',
        'addresses',
        'readonly_properties',
        'user_properties',
        'created',
        'modified',
    ]

    def change_view(self, request, object_id, form_url='', extra_context=None):
        extra_context = extra_context or {}
        link_model = self.model.source_link_set.field.model
        admin_url = 'admin:{0}_link_change'.format(self.opts.app_label)
        extra_context.update(
            {
                'node_links': link_model.objects.select_related('source', 'target')
                .only('source__label', 'target__label', 'cost', 'status')
                .filter(Q(source_id=object_id) | Q(target_id=object_id)),
                'admin_url': admin_url,
            }
        )
        return super().change_view(request, object_id, form_url, extra_context)


@admin.register(Link)
class LinkAdmin(NodeLinkMixin, BaseAdmin):
    model = Link
    form = UserPropertiesForm
    raw_id_fields = ['source', 'target']
    search_fields = [
        'source__label',
        'target__label',
        'source__addresses',
        'target__addresses',
        'properties',
    ]
    list_display = [
        '__str__',
        'organization',
        'topology',
        'status',
        'cost',
        'cost_text',
    ]
    list_filter = [
        'status',
        ('organization', MultitenantOrgFilter),
        ('topology', MultitenantRelatedOrgFilter),
    ]
    multitenant_shared_relations = ('topology', 'source', 'target')
    fields = [
        'topology',
        'organization',
        'status',
        'source',
        'target',
        'cost',
        'cost_text',
        'readonly_properties',
        'user_properties',
        'created',
        'modified',
    ]



# Generated by Django 3.0.7 on 2020-06-22 16:24

import collections

import jsonfield.fields
import rest_framework.utils.encoders
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('topology', '0009_remove_old_addresses'),
    ]

    operations = [
        migrations.AlterField(
            model_name='link',
            name='properties',
            field=jsonfield.fields.JSONField(
                blank=True,
                default=dict,
                dump_kwargs={
                    'cls': rest_framework.utils.encoders.JSONEncoder,
                    'indent': 4,
                },
                load_kwargs={'object_pairs_hook': collections.OrderedDict},
            ),
        ),
        migrations.AlterField(
            model_name='node',
            name='addresses',
            field=jsonfield.fields.JSONField(blank=True, default=[]),
        ),
        migrations.AlterField(
            model_name='node',
            name='properties',
            field=jsonfield.fields.JSONField(
                blank=True,
                default=dict,
                dump_kwargs={
                    'cls': rest_framework.utils.encoders.JSONEncoder,
                    'indent': 4,
                },
                load_kwargs={'object_pairs_hook': collections.OrderedDict},
            ),
        ),
    ]



from swapper import swappable_setting

from .base.models import AbstractDeviceNode


class DeviceNode(AbstractDeviceNode):
    class Meta(AbstractDeviceNode.Meta):
        abstract = False
        swappable = swappable_setting('topology_device', 'DeviceNode')



import swapper
from celery import shared_task


@shared_task
def create_device_node_relation(node_pk):
    Node = swapper.load_model('topology', 'Node')
    DeviceNode = swapper.load_model('topology_device', 'DeviceNode')
    node = Node.objects.select_related('topology').get(pk=node_pk)
    DeviceNode.auto_create(node)


@shared_task
def trigger_device_updates(link_pk):
    Link = swapper.load_model('topology', 'Link')
    DeviceNode = swapper.load_model('topology_device', 'DeviceNode')
    link = Link.objects.select_related('topology').get(pk=link_pk)
    DeviceNode.trigger_device_updates(link)



import os

from django.conf import settings
from django.conf.urls import include, url
from django.contrib import admin
from django.contrib.staticfiles.urls import staticfiles_urlpatterns

from openwisp_users.api.urls import get_api_urls

urlpatterns = []

if os.environ.get('SAMPLE_APP', False):
    from openwisp_network_topology.utils import get_visualizer_urls

    from .sample_network_topology.visualizer import views

    urlpatterns += [
        url(r'^topology/', include(get_visualizer_urls(views))),
    ]

urlpatterns += [
    url(r'^', include('openwisp_network_topology.urls')),
    url(r'^admin/', admin.site.urls),
    url(r'^api/v1/', include('openwisp_utils.api.urls')),
    url(r'^api/v1/', include(get_api_urls())),
]

urlpatterns += staticfiles_urlpatterns()


if settings.DEBUG and 'debug_toolbar' in settings.INSTALLED_APPS:
    import debug_toolbar

    urlpatterns += [url(r'^__debug__/', include(debug_toolbar.urls))]



import swapper
from django.test import TestCase
from rest_framework.utils.encoders import JSONEncoder

from .utils import CreateGraphObjectsMixin, CreateOrgMixin

Node = swapper.load_model('topology', 'Node')
Topology = swapper.load_model('topology', 'Topology')


class TestNode(CreateGraphObjectsMixin, CreateOrgMixin, TestCase):
    topology_model = Topology
    node_model = Node
    maxDiff = 0

    def setUp(self):
        org = self._create_org()
        t = self._create_topology(organization=org)
        self._create_node(
            label='node1', addresses=['192.168.0.1'], topology=t, organization=org
        )
        self._create_node(
            label='node2', addresses=['192.168.0.2'], topology=t, organization=org
        )

    def test_str(self):
        n = self.node_model(addresses=['192.168.0.1'], label='test node')
        self.assertIsInstance(str(n), str)

    def test_clean_properties(self):
        t = self.topology_model.objects.first()
        n = t._create_node(
            addresses=['192.168.0.1'], label='test node', properties=None
        )
        n.full_clean()
        self.assertEqual(n.properties, {})

    def test_node_address_single(self):
        t = self.topology_model.objects.first()
        n = t._create_node(
            addresses=['192.168.0.1'], label='test node', properties=None
        )
        n.full_clean()
        n.save()
        self.assertEqual(n.addresses[0], '192.168.0.1')
        self.assertEqual(n.addresses, ['192.168.0.1'])

    def test_node_address_multiple(self):
        t = self.topology_model.objects.first()
        n = t._create_node(
            label='test node',
            addresses=['192.168.0.1', '10.0.0.1', '10.0.0.2', '10.0.0.3'],
        )
        self.assertEqual(
            n.addresses, ['192.168.0.1', '10.0.0.1', '10.0.0.2', '10.0.0.3']
        )

    def test_node_local_addresses(self):
        t = self.topology_model.objects.first()
        n = t._create_node(
            label='test node', addresses=['192.168.0.1', '10.0.0.1', '10.0.0.2']
        )
        self.assertEqual(n.local_addresses, ['10.0.0.1', '10.0.0.2'])

    def test_node_name(self):
        t = self.topology_model.objects.first()
        n = t._create_node(addresses=['192.168.0.1', '10.0.0.1'])
        self.assertEqual(n.name, '192.168.0.1')
        n.label = 'test node'
        self.assertEqual(n.name, 'test node')

    def test_node_without_name(self):
        # According to the RFC, a node MAY not have name nor local_addresses defined
        t = self.topology_model.objects.first()
        n = t._create_node(addresses=[])
        self.assertEqual(n.name, '')

    def test_json(self):
        t = self.topology_model.objects.first()
        n = t._create_node(
            label='test node',
            addresses=['192.168.0.1', '10.0.0.1'],
            properties={"gateway": True},
        )
        self.assertEqual(
            dict(n.json(dict=True)),
            {
                'id': '192.168.0.1',
                'label': 'test node',
                'local_addresses': ['10.0.0.1'],
                'properties': {
                    'gateway': True,
                    'created': JSONEncoder().default(n.created),
                    'modified': JSONEncoder().default(n.modified),
                },
            },
        )
        self.assertIsInstance(n.json(), str)
        with self.subTest('testing original=True'):
            netjson = n.json(dict=True, original=True)
            self.assertNotIn('created', netjson['properties'])
            self.assertNotIn('modified', netjson['properties'])

    def test_get_from_address(self):
        t = self.topology_model.objects.first()
        n = t._create_node(addresses=['192.168.0.1', '10.0.0.1'])
        n.full_clean()
        n.save()
        self.assertIsInstance(
            self.node_model.get_from_address('192.168.0.1', t), self.node_model
        )
        self.assertIsInstance(
            self.node_model.get_from_address('10.0.0.1', t), self.node_model
        )
        self.assertIsNone(self.node_model.get_from_address('wrong', t))

    def test_count_address(self):
        t = self.topology_model.objects.first()
        self.assertEqual(self.node_model.count_address('192.168.0.1', t), 1)
        self.assertEqual(self.node_model.count_address('0.0.0.0', t), 0)

    def test_count_address_issue_58(self):
        t = self.topology_model.objects.first()
        n1 = t._create_node(addresses=['Benz_Kalloni'])
        n1.full_clean()
        n1.save()
        n2 = t._create_node(addresses=['Kalloni'])
        n2.full_clean()
        n2.save()
        self.assertEqual(self.node_model.count_address('Benz_Kalloni', t), 1)
        self.assertEqual(self.node_model.count_address('Kalloni', t), 1)

    def test_node_auto_org(self):
        t = self.topology_model.objects.first()
        n = self.node_model(
            label='TestNode', addresses=['192.168.0.1'], properties={}, topology=t
        )
        n.full_clean()
        self.assertEqual(n.organization, t.organization)

    def test_user_properties_in_json(self):
        t = self.topology_model.objects.first()
        n = t._create_node(
            addresses=['192.168.0.1'], label='test node', properties=None
        )
        n.properties = {
            'gateway': True,
        }
        n.user_properties = {'user_property': True}
        n.full_clean()
        n.save()

        with self.subTest('view json with original False'):
            data = n.json(dict=True)
            self.assertIn('gateway', data['properties'])
            self.assertIn('user_property', data['properties'])

        with self.subTest('view json with original True'):
            data = n.json(dict=True, original=True)
            self.assertIn('gateway', data['properties'])
            self.assertNotIn('user_property', data['properties'])



import json
import uuid

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group, Permission
from django.core.management import call_command
from django.core.management.base import BaseCommand
from swapper import get_model_name, load_model, split

User = get_user_model()
Topology = load_model('topology', 'Topology')
Organization = load_model('openwisp_users', 'Organization')


class BaseUpdateFromDjangoNetjsonGraph(BaseCommand):
    help = 'Upgrade from django-netjsongraph'
    app_label_users = split(get_model_name('openwisp_users', 'Organization'))[0]
    app_label = split(get_model_name('topology', 'Topology'))[0]

    def add_arguments(self, parser):
        parser.add_argument(
            '--backup',
            action='store',
            default=settings.BASE_DIR,
            help='(Optional) Path to the backup files',
        )
        parser.add_argument(
            '--organization',
            action='store',
            default=None,
            help=(
                '(Optional) organization UUID of the organization in '
                'which you want to import the data.'
            ),
        )

    def int_to_uuid(self, pk):
        return "00000000-0000-0000-0000-{:012}".format(pk)

    def _get_updated_permission_list(
        self, permission_data, permissions_list, contenttype_data
    ):
        permit_list = []
        for permit_pk in permissions_list:
            for item in permission_data:
                if item['pk'] == permit_pk:
                    for content in contenttype_data:
                        if item['fields']['content_type'] == content['pk']:
                            permit_app_label = content['fields']['app_label']
                            if permit_app_label == 'django_netjsongraph':
                                permit_app_label = self.app_label
                            elif (
                                content['fields']['model'] in ['user', 'group']
                                and permit_app_label == 'auth'
                            ):
                                permit_app_label = self.app_label_users
                    try:
                        permit_list.append(
                            Permission.objects.get(
                                content_type__app_label=permit_app_label,
                                codename=item['fields']['codename'],
                            ).pk
                        )
                    except Permission.DoesNotExist:  # pragma: nocover
                        pass
        return permit_list

    def handle(self, *args, **options):
        if options['organization']:
            org = Organization.objects.get(pk=options['organization'])
        else:
            org = Organization.objects.first()

        # Netjsongraph Models
        with open(f'{options["backup"]}/netjsongraph.json') as netjsongraph:
            netjsongraph_data = json.load(netjsongraph)
        # Make changes
        for data in netjsongraph_data:
            data['fields']['organization'] = str(org.pk)
            data['model'] = f'{self.app_label}.{data["model"].split(".")[1]}'
        # Save in anotherfile
        with open(f'{options["backup"]}/netjsongraph_loaded.json', 'w') as outfile:
            json.dump(netjsongraph_data, outfile)
        # Load to database
        call_command(
            'loaddata', f'{options["backup"]}/netjsongraph_loaded.json', verbosity=0
        )

        # Group Model
        with open(f'{options["backup"]}/contenttype.json') as contenttype:
            contenttype_data = json.load(contenttype)
        with open(f'{options["backup"]}/permission.json') as permission:
            permission_data = json.load(permission)
        with open(f'{options["backup"]}/group.json') as group:
            group_data = json.load(group)
        load_group_data = []
        for data in group_data:
            if not Group.objects.filter(name=data['fields']['name']).exists():
                load_group_data.append(
                    {
                        'model': f'{self.app_label_users}.group',
                        'pk': data['pk'] + Group.objects.count(),
                        'fields': {
                            'name': data['fields']['name'],
                            'permissions': self._get_updated_permission_list(
                                permission_data,
                                data['fields']['permissions'],
                                contenttype_data,
                            ),
                        },
                    }
                )
        if load_group_data:
            # Save in anotherfile
            with open(f'{options["backup"]}/group_loaded.json', 'w') as outfile:
                json.dump(load_group_data, outfile)
            # Load to database
            call_command(
                'loaddata', f'{options["backup"]}/group_loaded.json', verbosity=0
            )

        # User Model
        with open(f'{options["backup"]}/user.json') as users:
            users_data = json.load(users)
        # Make changes
        org_users_data = []
        load_users_data = []
        for data in users_data:
            data['model'] = f'{self.app_label_users}.user'
            data['pk'] = self.int_to_uuid(data['pk'])
            # If the user doesn't have an email, give them a
            # @example.com email but in openwisp-network-topology
            # email is UNIQUE.
            if not data['fields']['email']:
                data['fields']['email'] = f'{data["fields"]["username"]}@example.com'
            group_list = []
            for group_pk in data['fields']['groups']:
                for item in group_data:
                    if item['pk'] == group_pk:
                        group_list.append(
                            Group.objects.filter(name=item['fields']['name']).first().pk
                        )
            data['fields']['groups'] = group_list
            data['fields']['user_permissions'] = self._get_updated_permission_list(
                permission_data, data['fields']['user_permissions'], contenttype_data
            )
            if not User.objects.filter(email=data['fields']['email']):
                load_users_data.append(data)
                if not data['fields']['is_superuser']:
                    org_users_data.append(
                        {
                            'model': f'{self.app_label_users}.organizationuser',
                            'pk': str(uuid.uuid4()),
                            'fields': {
                                'created': data['fields']['date_joined'],
                                'modified': data['fields']['date_joined'],
                                'is_admin': False,
                                'user': data['pk'],
                                'organization': str(org.pk),
                            },
                        }
                    )
        load_users_data.extend(org_users_data)
        if load_users_data:
            # Save in anotherfile
            with open(f'{options["backup"]}/user_loaded.json', 'w') as outfile:
                json.dump(load_users_data, outfile)
            # Load to database
            call_command(
                'loaddata', f'{options["backup"]}/user_loaded.json', verbosity=0
            )

        self.stdout.write(self.style.SUCCESS('Migration Process Complete!'))


class Command(BaseUpdateFromDjangoNetjsonGraph):
    pass



import logging
from contextlib import contextmanager

logger = logging.getLogger(__name__)


@contextmanager
def log_failure(action, obj):
    try:
        yield
    except Exception as e:
        msg = 'Failed to perform {0} on {1}'.format(action, obj.__repr__())
        logger.exception(msg)
        print(
            '{0}: {1}\nSee error log for more '
            'information\n'.format(msg, e.__class__.__name__)
        )



from django.db import migrations

from . import migrate_openvpn_ids_0012


class Migration(migrations.Migration):
    dependencies = [
        ('topology', '0011_fix_link_properties'),
    ]

    operations = [
        migrations.RunPython(
            migrate_openvpn_ids_0012, reverse_code=migrations.RunPython.noop
        )
    ]



import json
from collections import OrderedDict
from datetime import datetime, timedelta

from django.core.exceptions import ValidationError
from django.db import models
from django.urls import reverse
from django.utils.functional import cached_property
from django.utils.module_loading import import_string
from django.utils.timezone import now
from django.utils.translation import ugettext_lazy as _
from netdiff import NetJsonParser, diff
from rest_framework.utils.encoders import JSONEncoder

from openwisp_users.mixins import OrgMixin
from openwisp_utils.base import KeyField, TimeStampedEditableModel

from ..contextmanagers import log_failure
from ..settings import PARSERS, TIMEOUT
from ..utils import print_info

STRATEGIES = (('fetch', _('FETCH')), ('receive', _('RECEIVE')))


class AbstractTopology(OrgMixin, TimeStampedEditableModel):

    label = models.CharField(_('label'), max_length=64)
    parser = models.CharField(
        _('format'),
        choices=PARSERS,
        max_length=128,
        help_text=_('Select topology format'),
    )
    strategy = models.CharField(
        _('strategy'), max_length=16, choices=STRATEGIES, default='fetch', db_index=True
    )
    # fetch strategy
    url = models.URLField(
        _('url'),
        blank=True,
        help_text=_('Topology data will be fetched from this URL' ' (FETCH strategy)'),
    )
    # receive strategy
    key = KeyField(
        unique=False,
        db_index=False,
        help_text=_('key needed to update topology from nodes '),
        verbose_name=_('key'),
        blank=True,
    )
    # receive strategy
    expiration_time = models.PositiveIntegerField(
        _('expiration time'),
        default=0,
        help_text=_(
            '"Expiration Time" in seconds: setting this to 0 will immediately mark missing links '
            'as down; a value higher than 0 will delay marking missing links as down until the '
            '"modified" field of a link is older than "Expiration Time"'
        ),
    )
    published = models.BooleanField(
        _('published'),
        default=True,
        help_text=_(
            'Unpublished topologies won\'t be updated or ' 'shown in the visualizer'
        ),
    )

    # the following fields will be filled automatically
    protocol = models.CharField(_('protocol'), max_length=64, blank=True)
    version = models.CharField(_('version'), max_length=24, blank=True)
    revision = models.CharField(_('revision'), max_length=64, blank=True)
    metric = models.CharField(_('metric'), max_length=24, blank=True)

    status = {'added': 'up', 'removed': 'down', 'changed': 'up'}
    action = {'added': 'add', 'changed': 'change', 'removed': 'remove'}

    class Meta:
        verbose_name_plural = _('topologies')
        abstract = True

    def __str__(self):
        return '{0} - {1}'.format(self.label, self.get_parser_display())

    def get_absolute_url(self):
        return reverse('topology_detail', args=[self.pk])

    def clean(self):
        if self.strategy == 'fetch' and not self.url:
            raise ValidationError(
                {'url': [_('an url must be specified when using FETCH strategy')]}
            )
        elif self.strategy == 'receive' and not self.key:
            raise ValidationError(
                {'key': [_('a key must be specified when using RECEIVE strategy')]}
            )

    @cached_property
    def parser_class(self):
        return import_string(self.parser)

    @property
    def link_model(self):
        return self.link_set.model

    @property
    def node_model(self):
        return self.node_set.model

    @property
    def snapshot_model(self):
        return self.snapshot_set.model

    def get_topology_data(self, data=None):
        """
        gets latest topology data
        """
        # if data is ``None`` it will be fetched from ``self.url``
        latest = self.parser_class(data=data, url=self.url, timeout=TIMEOUT)
        # update topology attributes if needed
        changed = False
        for attr in ['protocol', 'version', 'metric']:
            latest_attr = getattr(latest, attr)
            if getattr(self, attr) != latest_attr:
                setattr(self, attr, latest_attr)
                changed = True
        if changed:
            self.save()
        return latest

    def diff(self, data=None):
        """ shortcut to netdiff.diff """
        # if we get an instance of ``self.parser_class`` it means
        # ``self.get_topology_data`` has already been executed by ``receive``
        if isinstance(data, self.parser_class):
            latest = data
        else:
            latest = self.get_topology_data(data)
        current = NetJsonParser(self.json(dict=True, omit_down=True, original=True))
        return diff(current, latest)

    def get_nodes_queryset(self):
        return self.node_set.all()

    def get_links_queryset(self):
        return self.link_set.select_related('source', 'target')

    def json(self, dict=False, omit_down=False, original=False, **kwargs):
        """ returns a dict that represents a NetJSON NetworkGraph object """
        nodes = []
        links = []
        links_queryset = self.get_links_queryset()
        # needed to detect links coming back online
        if omit_down:
            links_queryset = links_queryset.filter(status='up')
        # populate graph
        for link in links_queryset:
            links.append(link.json(dict=True, original=original))
        for node in self.get_nodes_queryset():
            nodes.append(node.json(dict=True, original=original))
        netjson = OrderedDict(
            (
                ('type', 'NetworkGraph'),
                ('protocol', self.protocol),
                ('version', self.version),
                ('metric', self.metric),
                ('label', self.label),
                ('id', str(self.id)),
                ('parser', self.parser),
                ('created', self.created),
                ('modified', self.modified),
                ('nodes', nodes),
                ('links', links),
            )
        )
        if dict:
            return netjson
        return json.dumps(netjson, cls=JSONEncoder, **kwargs)

    def _create_node(self, **kwargs):
        options = dict(organization=self.organization, topology=self)
        options.update(kwargs)
        node = self.node_model(**options)
        return node

    def _create_link(self, **kwargs):
        options = dict(organization=self.organization, topology=self)
        options.update(kwargs)
        link = self.link_model(**options)
        return link

    def _update_added_items(self, items):
        Link, Node = self.link_model, self.node_model

        for node_dict in items.get('nodes', []):
            # if node exists, update its properties
            node = Node.get_from_address(node_dict['id'], topology=self)
            if node:
                self._update_node_properties(node, node_dict, section='added')
                continue
            # if node doesn't exist create new
            addresses = [node_dict['id']]
            addresses += node_dict.get('local_addresses', [])
            label = node_dict.get('label', '')
            properties = node_dict.get('properties', {})
            node = self._create_node(
                label=label, addresses=addresses, properties=properties
            )
            node.full_clean()
            node.save()

        for link_dict in items.get('links', []):
            link = Link.get_from_nodes(
                link_dict['source'], link_dict['target'], topology=self
            )
            # if link exists, update its properties
            if link:
                self._update_link_properties(link, link_dict, section='added')
                continue
            # if link does not exist create new
            source = Node.get_from_address(link_dict['source'], self)
            target = Node.get_from_address(link_dict['target'], self)
            link = self._create_link(
                source=source,
                target=target,
                cost=link_dict['cost'],
                cost_text=link_dict['cost_text'],
                properties=link_dict['properties'],
                topology=self,
            )
            link.full_clean()
            link.save()

    def _update_node_properties(self, node, node_dict, section):
        changed = False
        if node.label != node_dict.get('label'):
            changed = True
            node.label = node_dict.get('label')
        local_addresses = node_dict.get('local_addresses')
        if node.addresses != local_addresses:
            changed = True
            node.addresses = [node_dict['id']]
            if local_addresses:
                node.addresses += local_addresses
        if node.properties != node_dict.get('properties'):
            changed = True
            node.properties = node_dict.get('properties')
        # perform writes only if needed
        if changed:
            with log_failure(self.action[section], node):
                node.full_clean()
                node.save()

    def _update_link_properties(self, link, link_dict, section):
        changed = False
        # if status of link is changed
        if self.link_status_changed(link, self.status[section]):
            link.status = self.status[section]
            changed = True
        if link.cost != link_dict.get('cost'):
            link.cost = link_dict.get('cost')
            changed = True
        if link.cost_text != link_dict.get('cost_text'):
            link.cost_text = link_dict.get('cost_text')
            changed = True
        if link.properties != link_dict.get('properties'):
            link.properties = link_dict.get('properties')
            changed = True
        # perform writes only if needed
        if changed:
            with log_failure(self.action[section], link):
                link.full_clean()
                link.save()

    def _update_changed_items(self, items, section='changed'):
        Link, Node = self.link_model, self.node_model
        for node_dict in items.get('nodes', []):
            node = Node.get_from_address(node_dict['id'], topology=self)
            if node:
                self._update_node_properties(node, node_dict, section=section)

        for link_dict in items.get('links', []):
            link = Link.get_from_nodes(
                link_dict['source'], link_dict['target'], topology=self
            )
            if link:
                self._update_link_properties(link, link_dict, section=section)

    def update(self, data=None):
        """
        Updates topology
        Removed nodes are not deleted or modified
        Links are not deleted straightaway but set as "down"
        """
        diff = self.diff(data)

        if diff['added']:
            self._update_added_items(diff['added'])
        if diff['changed']:
            self._update_changed_items(diff['changed'])
        if diff['removed']:
            Link = self.link_model
            for link_dict in diff['removed'].get('links', []):
                link = Link.get_from_nodes(
                    link_dict['source'], link_dict['target'], topology=self
                )
                if link:
                    self._update_link_properties(link, link_dict, section='removed')

    def save_snapshot(self, **kwargs):
        """
        Saves the snapshot of topology
        """
        Snapshot = self.snapshot_model
        date = datetime.now().date()
        options = dict(organization=self.organization, topology=self, date=date)
        options.update(kwargs)
        try:
            s = Snapshot.objects.get(**options)
        except Snapshot.DoesNotExist:
            s = Snapshot(**options)
        s.data = self.json()
        s.save()

    def link_status_changed(self, link, status):
        """
        determines if link status has changed,
        takes in consideration also ``strategy`` and ``expiration_time``
        """
        status_changed = link.status != status
        # if status has not changed return ``False`` immediately
        if not status_changed:
            return False
        # if using fetch strategy or
        # using receive strategy and link is coming back up or
        # receive strategy and ``expiration_time == 0``
        elif self.strategy == 'fetch' or status == 'up' or self.expiration_time == 0:
            return True
        # if using receive strategy and expiration_time of link has expired
        elif link.modified < (now() - timedelta(seconds=self.expiration_time)):
            return True
        # if using receive strategy and expiration_time of link has not expired
        return False

    def receive(self, data):
        """
        Receive topology data (RECEIVE strategy)
        expiration_time at 0 means:
          "if a link is missing, mark it as down immediately"
        expiration_time > 0 means:
          "if a link is missing, wait expiration_time seconds before marking it as down"
        """
        if self.expiration_time > 0:
            data = self.get_topology_data(data)
            Link = self.link_model
            netjson = data.json(dict=True)
            # update last modified date of all received links
            for link_dict in netjson['links']:
                link = Link.get_from_nodes(
                    link_dict['source'], link_dict['target'], topology=self
                )
                if link:
                    link.save()
        self.update(data)

    @classmethod
    def update_all(cls, label=None):
        """
        - updates topologies
        - logs failures
        - calls delete_expired_links()
        """
        queryset = cls.objects.filter(published=True, strategy='fetch')
        if label:
            queryset = queryset.filter(label__icontains=label)
        for topology in queryset:
            print_info('Updating topology {0}'.format(topology))
            with log_failure('update', topology):
                topology.update()
        cls().link_model.delete_expired_links()
        cls().node_model.delete_expired_nodes()

    @classmethod
    def save_snapshot_all(cls, label=None):
        """
        - save snapshots of topoogies
        - logs failures
        """
        queryset = cls.objects.filter(published=True)
        if label:
            queryset = queryset.filter(label__icontains=label)
        for topology in queryset:
            print_info('Saving topology {0}'.format(topology))
            with log_failure('save_snapshot', topology):
                topology.save_snapshot()



import re

import responses
import swapper
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission
from django.test import TestCase
from django.urls import reverse

from openwisp_users.tests.utils import TestMultitenantAdminMixin, TestOrganizationMixin

from ..admin import TopologyAdmin
from .utils import CreateGraphObjectsMixin, CreateOrgMixin, LoadMixin

Link = swapper.load_model('topology', 'Link')
Node = swapper.load_model('topology', 'Node')
Topology = swapper.load_model('topology', 'Topology')


class TestAdmin(CreateGraphObjectsMixin, CreateOrgMixin, LoadMixin, TestCase):
    module = 'openwisp_network_topology'
    app_label = 'topology'
    topology_model = Topology
    link_model = Link
    node_model = Node
    user_model = get_user_model()
    fixtures = ['test_users.json']
    api_urls_path = 'api.urls'

    @property
    def prefix(self):
        return 'admin:{0}'.format(self.app_label)

    def setUp(self):
        org = self._create_org()
        t = self._create_topology(organization=org)
        self._create_node(
            label='node1', addresses=['192.168.0.1'], topology=t, organization=org
        )
        self._create_node(
            label='node2', addresses=['192.168.0.2'], topology=t, organization=org
        )
        self.client.force_login(self.user_model.objects.get(username='admin'))
        self.changelist_path = reverse('{0}_topology_changelist'.format(self.prefix))

    def test_unpublish_selected(self):
        t = self.topology_model.objects.first()
        self.assertEqual(t.published, True)
        self.client.post(
            self.changelist_path,
            {'action': 'unpublish_selected', '_selected_action': str(t.pk)},
        )
        t.refresh_from_db()
        self.assertEqual(t.published, False)

    def test_publish_selected(self):
        t = self.topology_model.objects.first()
        t.published = False
        t.save()
        self.client.post(
            self.changelist_path,
            {'action': 'publish_selected', '_selected_action': str(t.pk)},
        )
        t.refresh_from_db()
        self.assertEqual(t.published, True)

    @responses.activate
    def test_update_selected(self):
        t = self.topology_model.objects.first()
        t.parser = 'netdiff.NetJsonParser'
        t.save()
        responses.add(
            responses.GET,
            'http://127.0.0.1:9090',
            body=self._load('static/netjson-1-link.json'),
            content_type='application/json',
        )
        self.node_model.objects.all().delete()
        self.client.post(
            self.changelist_path,
            {'action': 'update_selected', '_selected_action': str(t.pk)},
        )
        self.assertEqual(self.node_model.objects.count(), 2)
        self.assertEqual(self.link_model.objects.count(), 1)

    @responses.activate
    def test_update_selected_failed(self):
        t = self.topology_model.objects.first()
        t.parser = 'netdiff.NetJsonParser'
        t.save()
        responses.add(
            responses.GET,
            'http://127.0.0.1:9090',
            body='{"error": "not found"}',
            status=404,
            content_type='application/json',
        )
        self.node_model.objects.all().delete()
        response = self.client.post(
            self.changelist_path,
            {'action': 'update_selected', '_selected_action': str(t.pk)},
            follow=True,
        )
        self.assertEqual(self.node_model.objects.count(), 0)
        self.assertEqual(self.link_model.objects.count(), 0)
        message = list(response.context['messages'])[0]
        self.assertEqual(message.tags, 'error')
        self.assertIn('not updated', message.message)

    def test_topology_viewonsite(self):
        topology = self.topology_model.objects.first()
        path = reverse('{0}_topology_change'.format(self.prefix), args=[topology.pk])
        response = self.client.get(path)
        self.assertContains(response, 'View on site')
        # Pattern for the link
        pattern = '{0}{1}'.format(r'/admin/r/[0-9][0-9]?/', f'{topology.pk}')
        self.assertTrue(bool(re.compile(pattern).search(str(response.content))))

    def test_topology_receive_url(self):
        t = self.topology_model.objects.first()
        t.strategy = 'receive'
        t.save()
        path = reverse('{0}_topology_change'.format(self.prefix), args=[t.pk])
        response = self.client.get(path)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'field-receive_url')

    def test_custom_topology_receive_url(self):
        t = self.topology_model.objects.first()
        t.strategy = 'receive'
        t.save()
        path = reverse('{0}_topology_change'.format(self.prefix), args=[t.pk])
        # No change in URL Test
        receive_path = f'topology/{t.pk}/receive/'
        response = self.client.get(path)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'field-receive_url')
        self.assertContains(response, f'http://testserver/api/v1/{receive_path}')
        # Change URL Test
        TopologyAdmin.receive_url_baseurl = 'http://changedurlbase'
        response = self.client.get(path)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'field-receive_url')
        self.assertContains(response, f'http://changedurlbase/api/v1/{receive_path}')
        # Change URLConf Test
        TopologyAdmin.receive_url_urlconf = '{}.{}'.format(
            self.module, self.api_urls_path
        )
        response = self.client.get(path)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'field-receive_url')
        self.assertContains(response, f'http://changedurlbase/{receive_path}')
        # Reset test options
        TopologyAdmin.receive_url_baseurl = None
        TopologyAdmin.receive_url_urlconf = None

    def test_node_change_form(self):
        n = self.node_model.objects.first()
        path = reverse('{0}_node_change'.format(self.prefix), args=[n.pk])
        response = self.client.get(path)
        self.assertContains(response, 'Links to other nodes')
        self.assertNotContains(response, 'organization_id')
        self.assertContains(response, n.topology.organization.name)

    def test_node_change_list_queries(self):
        path = reverse('{0}_node_changelist'.format(self.prefix))
        with self.assertNumQueries(7):
            self.client.get(path)

    def test_link_change_list_queries(self):
        t = Topology.objects.first()
        n1 = self._create_node(label='node1org1', topology=t)
        n2 = self._create_node(label='node2org1', topology=t)
        self._create_link(topology=t, source=n1, target=n2)
        path = reverse('{0}_link_changelist'.format(self.prefix))
        with self.assertNumQueries(7):
            self.client.get(path)

    def test_link_change_form(self):
        t = Topology.objects.first()
        n1 = self._create_node(label='node1org1', topology=t)
        n2 = self._create_node(label='node2org1', topology=t)
        link = self._create_link(topology=t, source=n1, target=n2)
        path = reverse('{0}_link_change'.format(self.prefix), args=[link.pk])
        response = self.client.get(path)
        self.assertNotContains(response, 'organization_id')
        self.assertContains(response, link.topology.organization.name)

    def test_node_add(self):
        path = reverse('{0}_node_add'.format(self.prefix))
        response = self.client.get(path)
        self.assertNotContains(response, 'Links to other nodes')

    def test_topology_change_form(self):
        topology = self.topology_model.objects.first()
        path = reverse('{0}_topology_change'.format(self.prefix), args=[topology.pk])
        response = self.client.get(path)
        self.assertContains(response, 'View topology graph')
        self.assertContains(
            response, '<div class="readonly">{0}</div>'.format(topology.pk)
        )

    def test_topology_visualize_view(self):
        t = self.topology_model.objects.first()
        path = reverse('{0}_topology_visualize'.format(self.prefix), args=[t.pk])
        response = self.client.get(path)
        self.assertContains(response, 'd3.netJsonGraph(')

    def test_update_selected_receive_topology(self):
        t = self.topology_model.objects.first()
        t.label = 'test receive'
        t.parser = 'netdiff.NetJsonParser'
        t.strategy = 'receive'
        t.save()
        response = self.client.post(
            self.changelist_path,
            {'action': 'update_selected', '_selected_action': str(t.pk)},
            follow=True,
        )
        message = list(response.context['messages'])[0]
        self.assertEqual('warning', message.tags)
        self.assertIn('1 topology was ignored', message.message)

    def _test_properties_field(self, model, obj):
        with self.subTest('test old properties readonly'):
            content = 'readonly_properties">'
            r = self.client.get(reverse(f'{self.prefix}_{model}_add'))
            self.assertEqual(r.status_code, 200)
            self.assertContains(r, content)

        with self.subTest('test old properties display in list'):
            path = reverse(f'{self.prefix}_{model}_change', args=[obj.pk])
            r = self.client.get(path)
            self.assertEqual(r.status_code, 200)
            content = '<p><strong>Gateway</strong>: False</p>'
            self.assertContains(r, content)

        with self.subTest('test user properties diplay in flat json widgets'):
            r = self.client.get(reverse(f'{self.prefix}_{model}_add'))
            self.assertEqual(r.status_code, 200)
            self.assertContains(r, 'flat-json-user_properties')

    def test_node_user_properties_field(self):
        t = Topology.objects.first()
        n = self._create_node(label='node1org1', topology=t)
        n.properties = {
            'gateway': False,
        }
        n.full_clean()
        n.save()
        self._test_properties_field('node', n)

    def test_link_user_properties_field(self):
        t = Topology.objects.first()
        n1 = self._create_node(label='node1org1', topology=t)
        n2 = self._create_node(label='node2org1', topology=t)
        li = self._create_link(topology=t, source=n1, target=n2)
        li.properties = {
            'gateway': False,
        }
        li.full_clean()
        li.save()
        self._test_properties_field('link', li)


class TestMultitenantAdmin(
    CreateGraphObjectsMixin, TestMultitenantAdminMixin, TestOrganizationMixin, TestCase
):
    app_label = 'topology'
    topology_model = Topology
    node_model = Node
    link_model = Link

    operator_permission_filters = [
        {'codename__endswith': 'topology'},
        {'codename__endswith': 'node'},
        {'codename__endswith': 'link'},
    ]

    def _create_multitenancy_test_env(self):
        org1 = self._create_org(name='test1org')
        org2 = self._create_org(name='test2org')
        inactive = self._create_org(name='inactive-org', is_active=False)
        operator = self._create_operator(organizations=[org1, inactive])
        t1 = self._create_topology(label='topology1org', organization=org1)
        t2 = self._create_topology(label='topology2org', organization=org2)
        t3 = self._create_topology(label='topology3org', organization=inactive)
        n11 = self._create_node(label='node1org1', topology=t1, organization=org1)
        n12 = self._create_node(label='node2org1', topology=t1, organization=org1)
        n21 = self._create_node(label='node1org2', topology=t2, organization=org2)
        n22 = self._create_node(label='node2org2', topology=t2, organization=org2)
        n31 = self._create_node(
            label='node1inactive', topology=t3, organization=inactive
        )
        n32 = self._create_node(
            label='node2inactive', topology=t3, organization=inactive
        )
        l1 = self._create_link(topology=t1, organization=org1, source=n11, target=n12)
        l2 = self._create_link(topology=t2, organization=org2, source=n21, target=n22)
        l3 = self._create_link(
            topology=t3, organization=inactive, source=n31, target=n32
        )
        data = dict(
            t1=t1,
            t2=t2,
            t3_inactive=t3,
            n11=n11,
            n12=n12,
            l1=l1,
            n21=n21,
            n22=n22,
            l2=l2,
            n31=n31,
            n32=n32,
            l3_inactive=l3,
            org1=org1,
            org2=org2,
            inactive=inactive,
            operator=operator,
        )
        return data

    def test_topology_queryset(self):
        data = self._create_multitenancy_test_env()
        perm = Permission.objects.get_by_natural_key(
            'delete_topology', self.app_label, self.topology_model.__name__.lower()
        )
        data['operator'].user_permissions.remove(perm)
        self._test_multitenant_admin(
            url=reverse(f'admin:{self.app_label}_topology_changelist'),
            visible=[data['t1'].label, data['org1'].name],
            hidden=[data['t2'].label, data['org2'].name, data['t3_inactive'].label],
        )

    def test_topology_organization_fk_queryset(self):
        data = self._create_multitenancy_test_env()
        self._test_multitenant_admin(
            url=reverse(f'admin:{self.app_label}_topology_add'),
            visible=[data['org1'].name],
            hidden=[data['org2'].name, data['inactive']],
            select_widget=True,
        )

    def test_node_queryset(self):
        data = self._create_multitenancy_test_env()
        self._test_multitenant_admin(
            url=reverse(f'admin:{self.app_label}_node_changelist'),
            visible=[data['n11'].label, data['n12'].label, data['org1'].name],
            hidden=[
                data['n21'].label,
                data['n22'].label,
                data['org2'].name,
                data['n31'].label,
                data['n32'].label,
                data['inactive'],
            ],
        )

    def test_node_organization_fk_queryset(self):
        self._create_multitenancy_test_env()
        self._login(username='operator', password='tester')
        response = self.client.get(reverse(f'admin:{self.app_label}_node_add'))
        self.assertNotContains(response, 'organization_id')

    def test_link_queryset(self):
        data = self._create_multitenancy_test_env()
        self._test_multitenant_admin(
            url=reverse(f'admin:{self.app_label}_link_changelist'),
            visible=[str(data['l1']), data['org1'].name],
            hidden=[str(data['l2']), data['org2'].name, str(data['l3_inactive'])],
        )

    def test_link_organization_fk_queryset(self):
        self._create_multitenancy_test_env()
        self._login(username='operator', password='tester')
        response = self.client.get(reverse(f'admin:{self.app_label}_link_add'))
        self.assertNotContains(response, 'organization_id')

    def test_node_topology_fk_queryset(self):
        data = self._create_multitenancy_test_env()
        self._test_multitenant_admin(
            url=reverse(f'admin:{self.app_label}_node_add'),
            visible=[data['t1'].label],
            hidden=[data['t2'].label, data['t3_inactive'].label],
        )

    def test_link_topology_fk_queryset(self):
        data = self._create_multitenancy_test_env()
        self._test_multitenant_admin(
            url=reverse(f'admin:{self.app_label}_link_add'),
            visible=[data['t1'].label],
            hidden=[data['t2'].label, data['t3_inactive'].label],
        )

    def test_node_topology_filter(self):
        data = self._create_multitenancy_test_env()
        t_special = self._create_topology(label='special', organization=data['org1'])
        self._test_multitenant_admin(
            url=reverse(f'admin:{self.app_label}_node_changelist'),
            visible=[data['t1'].label, t_special.label],
            hidden=[data['t2'].label, data['t3_inactive'].label],
        )

    def test_link_topology_filter(self):
        data = self._create_multitenancy_test_env()
        t_special = self._create_topology(label='special', organization=data['org1'])
        self._test_multitenant_admin(
            url=reverse(f'admin:{self.app_label}_link_changelist'),
            visible=[data['t1'].label, t_special.label],
            hidden=[data['t2'].label, data['t3_inactive'].label],
        )



# Generated by Django 2.2.14 on 2020-07-15 21:25

import uuid

import django.db.models.deletion
import swapper
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        swapper.dependency('topology', 'Node'),
        swapper.dependency('config', 'Device'),
    ]

    operations = [
        migrations.CreateModel(
            name='DeviceNode',
            fields=[
                (
                    'id',
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    'device',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=swapper.get_model_name('config', 'Device'),
                    ),
                ),
                (
                    'node',
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=swapper.get_model_name('topology', 'Node'),
                    ),
                ),
            ],
            options={
                'abstract': False,
                'swappable': swapper.swappable_setting('topology_device', 'DeviceNode'),
                'unique_together': {('node', 'device')},
            },
        ),
    ]



import json
import sys
from contextlib import contextmanager
from datetime import timedelta
from io import StringIO

import responses
import swapper
from django.core.management import call_command
from django.test import TestCase
from django.utils.timezone import now

from openwisp_users.models import Organization

from .. import settings as app_settings
from .utils import CreateGraphObjectsMixin, CreateOrgMixin, LoadMixin

Link = swapper.load_model('topology', 'Link')
Node = swapper.load_model('topology', 'Node')
Snapshot = swapper.load_model('topology', 'Snapshot')
Topology = swapper.load_model('topology', 'Topology')


@contextmanager
def redirect_stdout(stream):
    sys.stdout = stream
    try:
        yield
    finally:
        sys.stdout = sys.__stdout__


class TestUtils(CreateGraphObjectsMixin, CreateOrgMixin, LoadMixin, TestCase):
    """
    tests for django_netjsongraph.utils
    """

    topology_model = Topology
    node_model = Node
    link_model = Link
    snapshot_model = Snapshot
    organization_model = Organization
    maxDiff = None

    def setUp(self):
        org = self._create_org()
        t = self._create_topology(organization=org)
        self._create_node(
            label='node1', addresses=['192.168.0.1'], topology=t, organization=org
        )
        self._create_node(
            label='node2', addresses=['192.168.0.2'], topology=t, organization=org
        )

    def _create_link(self, **kwargs):
        org = self.organization_model.objects.first()
        options = dict(cost='1', cost_text='one', properties={}, organization=org)
        options.update(kwargs)
        link = self.link_model(**options)
        link.full_clean()
        link.save()
        return link

    @responses.activate
    def test_update_all_method(self):
        t = self.topology_model.objects.first()
        t.parser = 'netdiff.NetJsonParser'
        t.save()
        responses.add(
            responses.GET,
            'http://127.0.0.1:9090',
            body=self._load('static/netjson-1-link.json'),
            content_type='application/json',
        )
        self.node_model.objects.all().delete()
        self.topology_model.update_all('testnetwork')
        self.assertEqual(self.node_model.objects.count(), 2)
        self.assertEqual(self.link_model.objects.count(), 1)
        # test exception
        t.url = t.url.replace('9090', '9091')
        t.save()
        self.node_model.objects.all().delete()
        self.link_model.objects.all().delete()
        responses.add(
            responses.GET,
            'http://127.0.0.1:9091',
            body=self._load('static/netjson-invalid.json'),
            content_type='application/json',
        )
        # capture output
        output = StringIO()
        with redirect_stdout(output):
            self.topology_model.update_all()

        self.assertEqual(self.node_model.objects.count(), 1)
        self.assertEqual(self.link_model.objects.count(), 0)
        self.assertIn('Failed to', output.getvalue())

    @responses.activate
    def test_update_topology_command(self):
        t = self.topology_model.objects.first()
        t.parser = 'netdiff.NetJsonParser'
        t.save()
        responses.add(
            responses.GET,
            'http://127.0.0.1:9090',
            body=self._load('static/netjson-1-link.json'),
            content_type='application/json',
        )
        self.node_model.objects.all().delete()
        self.topology_model.update_all()
        self.assertEqual(self.node_model.objects.count(), 2)
        self.assertEqual(self.link_model.objects.count(), 1)
        # test exception
        t.url = t.url.replace('9090', '9091')
        t.save()
        self.node_model.objects.all().delete()
        self.link_model.objects.all().delete()
        responses.add(
            responses.GET,
            'http://127.0.0.1:9091',
            body=self._load('static/netjson-invalid.json'),
            content_type='application/json',
        )
        # capture output
        output = StringIO()
        with redirect_stdout(output):
            call_command('update_topology')

        self.assertEqual(self.node_model.objects.count(), 1)
        self.assertEqual(self.link_model.objects.count(), 0)
        self.assertIn('Failed to', output.getvalue())

    @responses.activate
    def test_update_all_method_unpublished(self):
        t = self.topology_model.objects.first()
        t.published = False
        t.parser = 'netdiff.NetJsonParser'
        t.save()
        responses.add(
            responses.GET,
            'http://127.0.0.1:9090',
            body=self._load('static/netjson-1-link.json'),
            content_type='application/json',
        )
        self.node_model.objects.all().delete()
        self.topology_model.update_all()
        self.assertEqual(self.node_model.objects.count(), 0)
        self.assertEqual(self.link_model.objects.count(), 0)

    @responses.activate
    def test_delete_expired_links(self):
        t = self.topology_model.objects.first()
        t.parser = 'netdiff.NetJsonParser'
        t.save()
        # should not delete
        almost_expired_date = now() - timedelta(days=app_settings.LINK_EXPIRATION - 10)
        n1 = self.node_model.objects.all()[0]
        n2 = self.node_model.objects.all()[1]
        link = self._create_link(
            source=n1, target=n2, cost=1, status='down', topology=t
        )
        self.link_model.objects.filter(pk=link.pk).update(
            created=almost_expired_date, modified=almost_expired_date
        )
        empty_topology = json.dumps(
            {
                "type": "NetworkGraph",
                "protocol": "OLSR",
                "version": "0.8",
                "metric": "ETX",
                "nodes": [],
                "links": [],
            }
        )
        responses.add(
            responses.GET,
            'http://127.0.0.1:9090',
            body=empty_topology,
            content_type='application/json',
        )
        self.topology_model.update_all('testnetwork')
        self.assertEqual(self.node_model.objects.count(), 2)
        self.assertEqual(self.link_model.objects.count(), 1)
        # should delete
        expired_date = now() - timedelta(days=app_settings.LINK_EXPIRATION + 10)
        self.link_model.objects.filter(pk=link.pk).update(
            created=expired_date, modified=expired_date
        )
        self.topology_model.update_all('testnetwork')
        self.assertEqual(self.node_model.objects.count(), 2)
        self.assertEqual(self.link_model.objects.count(), 0)

    @responses.activate
    def test_delete_expired_nodes(self):
        NODE_EXPIRATION = getattr(app_settings, 'NODE_EXPIRATION')
        # Test with the default value(False)
        # Should not delete
        setattr(app_settings, 'NODE_EXPIRATION', False)
        t = self.topology_model.objects.first()
        t.parser = 'netdiff.NetJsonParser'
        t.save()
        expired_date = now() - timedelta(days=60)
        n1 = self.node_model.objects.all()[0]
        n2 = self.node_model.objects.all()[1]
        self.node_model.objects.filter(pk=n1.pk).update(
            created=expired_date, modified=expired_date
        )
        self.node_model.objects.filter(pk=n2.pk).update(
            created=expired_date, modified=expired_date
        )
        empty_topology = json.dumps(
            {
                "type": "NetworkGraph",
                "protocol": "OLSR",
                "version": "0.8",
                "metric": "ETX",
                "nodes": [],
                "links": [],
            }
        )
        responses.add(
            responses.GET,
            'http://127.0.0.1:9090',
            body=empty_topology,
            content_type='application/json',
        )
        self.topology_model.update_all('testnetwork')
        self.assertEqual(self.node_model.objects.count(), 2)

        # Test with a custom value
        # Should delete
        setattr(app_settings, 'NODE_EXPIRATION', 60)
        expired_date = now() - timedelta(days=app_settings.NODE_EXPIRATION + 10)
        self.node_model.objects.filter(pk=n1.pk).update(
            created=expired_date, modified=expired_date
        )
        self.node_model.objects.filter(pk=n2.pk).update(
            created=expired_date, modified=expired_date
        )
        self.topology_model.update_all('testnetwork')
        self.assertEqual(self.node_model.objects.count(), 0)
        self.assertEqual(self.link_model.objects.count(), 0)
        # Set the setting to it's original value
        setattr(app_settings, 'NODE_EXPIRATION', NODE_EXPIRATION)

    @responses.activate
    def test_delete_expired_disabled(self):
        t = self.topology_model.objects.first()
        t.parser = 'netdiff.NetJsonParser'
        t.save()
        n1 = self.node_model.objects.all()[0]
        n2 = self.node_model.objects.all()[1]
        link = self._create_link(
            source=n1, target=n2, cost=1, status='down', topology=t
        )
        expired_date = now() - timedelta(days=app_settings.LINK_EXPIRATION + 10)
        self.link_model.objects.filter(pk=link.pk).update(
            created=expired_date, modified=expired_date
        )
        empty_topology = json.dumps(
            {
                "type": "NetworkGraph",
                "protocol": "OLSR",
                "version": "0.8",
                "metric": "ETX",
                "nodes": [],
                "links": [],
            }
        )
        responses.add(
            responses.GET,
            'http://127.0.0.1:9090',
            body=empty_topology,
            content_type='application/json',
        )
        ORIGINAL_LINK_EXPIRATION = int(app_settings.LINK_EXPIRATION)
        app_settings.LINK_EXPIRATION = False
        self.topology_model.update_all('testnetwork')
        self.assertEqual(self.node_model.objects.count(), 2)
        self.assertEqual(self.link_model.objects.count(), 1)
        app_settings.LINK_EXPIRATION = ORIGINAL_LINK_EXPIRATION

    def test_save_snapshot_all_method(self, **kwargs):
        org = self._create_org()
        options = dict(organization=org)
        options.update(**kwargs)
        self.assertEqual(self.snapshot_model.objects.count(), 0)
        self.topology_model.save_snapshot_all('testnetwork')
        self.assertEqual(self.snapshot_model.objects.count(), 1)
        self._create_topology(**options)
        self.topology_model.save_snapshot_all()
        self.assertEqual(self.snapshot_model.objects.count(), 2)

    def test_save_snapshot_command(self):
        self.assertEqual(self.snapshot_model.objects.count(), 0)
        output = StringIO()
        with redirect_stdout(output):
            call_command('save_snapshot')
        self.assertEqual(self.snapshot_model.objects.count(), 1)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.3 on 2017-08-15 10:01

import uuid

import django.db.models.deletion
import django.utils.timezone
import model_utils.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('topology', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Snapshot',
            fields=[
                (
                    'id',
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    'created',
                    model_utils.fields.AutoCreatedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name='created',
                    ),
                ),
                (
                    'modified',
                    model_utils.fields.AutoLastModifiedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name='modified',
                    ),
                ),
                ('data', models.TextField()),
                ('date', models.DateField(auto_now=True)),
                (
                    'organization',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to='openwisp_users.Organization',
                        verbose_name='organization',
                    ),
                ),
                (
                    'topology',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to='topology.Topology',
                    ),
                ),
            ],
            options={'abstract': False, 'verbose_name_plural': 'snapshots'},
        ),
    ]



# Manually written, used during migrations
import swapper


def get_model(apps, app_name, model):
    model_name = swapper.get_model_name(app_name, model)
    model_label = swapper.split(model_name)[0]
    return apps.get_model(model_label, model)


def migrate_addresses(apps, schema_editor):
    Node = get_model(apps, 'topology', 'Node')
    for node in Node.objects.iterator():
        addresses = node.addresses_old.replace(' ', '')
        if addresses.startswith(';'):
            addresses = addresses[1:]
        addresses = addresses[0:-1].split(';')
        node.addresses = addresses
        node.save()


def migrate_openvpn_ids_0012(apps, schema_editor):
    Node = get_model(apps, 'topology', 'Node')
    queryset = Node.objects.filter(topology__parser='netdiff.OpenvpnParser')
    for node in queryset.iterator():
        node.addresses[0] = node.label
        node.full_clean()
        node.save()


def fix_link_properties(apps, schema_editor):
    Link = get_model(apps, 'topology', 'Link')
    for link in Link.objects.all():
        link.full_clean()
        link.save()



import swapper
from django.test import TestCase

Snapshot = swapper.load_model('topology', 'Snapshot')


class TestSnapshot(TestCase):
    snapshot_model = Snapshot

    def test_str(self):
        s = self.snapshot_model.objects.first()
        self.assertIsInstance(str(s), str)



from openwisp_network_topology.visualizer.views import (
    TopologyDetailView as BaseTopologyDetailView,
)
from openwisp_network_topology.visualizer.views import (
    TopologyListView as BaseTopologyListView,
)


class TopologyListView(BaseTopologyListView):
    pass


class TopologyDetailView(BaseTopologyDetailView):
    pass


topology_list = TopologyListView.as_view()
topology_detail = TopologyDetailView.as_view()



import logging
import os

from openwisp_users.models import Organization
from openwisp_utils.tests import TimeLoggingTestRunner


class CreateOrgMixin(object):
    def _create_org(self, **kwargs):
        options = dict(name='test-organization')
        options.update(kwargs)
        org = Organization(**options)
        org.save()
        return org


class CreateGraphObjectsMixin(object):
    def _create_topology(self, **kwargs):
        options = dict(
            label='TestNetwork',
            parser='netdiff.OlsrParser',
            strategy='fetch',
            url='http://127.0.0.1:9090',
            protocol='OLSR',
            version='0.8',
            metric='ETX',
            created='2017-07-10T20:02:52.483Z',
            modified='2015-07-14T20:02:52.483Z',
        )
        options.update(kwargs)
        t = self.topology_model(**options)
        t.full_clean()
        t.save()
        return t

    def _create_node(self, **kwargs):
        options = dict(
            label='TestNode',
            addresses=['192.168.0.1'],
            created='2017-07-10T20:02:52.483Z',
            modified='2017-07-14T20:02:52.483Z',
            properties={},
        )
        options.update(kwargs)
        n = self.node_model(**options)
        n.full_clean()
        n.save()
        return n

    def _create_link(self, **kwargs):
        options = dict(cost='1', cost_text='one', properties={})
        options.update(kwargs)
        link = self.link_model(**options)
        link.full_clean()
        link.save()
        return link


class LoggingDisabledTestRunner(TimeLoggingTestRunner):
    def run_tests(self, test_labels, extra_tests=None, **kwargs):
        # disable logging below CRITICAL while testing
        logging.disable(logging.CRITICAL)
        return super().run_tests(test_labels, extra_tests, **kwargs)


class UnpublishMixin(object):
    def _unpublish(self):
        t = self.topology_model.objects.first()
        t.published = False
        t.save()


class LoadMixin(object):
    def _load(self, file):
        d = os.path.dirname(os.path.abspath(__file__))
        return open(os.path.join(d, file)).read()



# RENAME THIS FILE TO local_settings.py IF YOU NEED TO CUSTOMIZE SOME SETTINGS
# BUT DO NOT COMMIT

# DATABASES = {
#    'default': {
#        'ENGINE': 'django.db.backends.sqlite3',
#        'NAME': 'openwisp_network_topology.db',
#        'USER': '',
#        'PASSWORD': '',
#        'HOST': '',
#        'PORT': ''
#    },
# }



import swapper
from django.apps import AppConfig
from django.conf import settings
from django.utils.translation import ugettext_lazy as _

from .settings import SIGNALS


class OpenwispNetworkTopologyConfig(AppConfig):
    name = 'openwisp_network_topology'
    label = 'topology'
    verbose_name = _('Network Topology')

    def ready(self, *args, **kwargs):
        if SIGNALS:  # pragma: nocover
            __import__(SIGNALS)
        self.add_default_menu_items()

    def add_default_menu_items(self):
        menu_setting = 'OPENWISP_DEFAULT_ADMIN_MENU_ITEMS'
        items = [{'model': swapper.get_model_name('topology', 'Topology')}]
        if not hasattr(settings, menu_setting):
            setattr(settings, menu_setting, items)
        else:
            current_menu = getattr(settings, menu_setting)
            current_menu += items



from openwisp_network_topology.apps import OpenwispNetworkTopologyConfig


class SampleNetworkTopologyConfig(OpenwispNetworkTopologyConfig):
    name = 'openwisp2.sample_network_topology'
    label = 'sample_network_topology'



import json
from collections import OrderedDict
from datetime import timedelta

import swapper
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Q
from django.utils.timezone import now
from django.utils.translation import ugettext_lazy as _
from jsonfield import JSONField
from model_utils import Choices
from model_utils.fields import StatusField
from rest_framework.utils.encoders import JSONEncoder

from openwisp_users.mixins import OrgMixin
from openwisp_utils.base import TimeStampedEditableModel

from .. import settings as app_settings
from ..utils import link_status_changed, print_info


class AbstractLink(OrgMixin, TimeStampedEditableModel):
    """
    NetJSON NetworkGraph Link Object implementation
    """

    topology = models.ForeignKey(
        swapper.get_model_name('topology', 'Topology'), on_delete=models.CASCADE
    )
    source = models.ForeignKey(
        swapper.get_model_name('topology', 'Node'),
        related_name='source_link_set',
        on_delete=models.CASCADE,
    )
    target = models.ForeignKey(
        swapper.get_model_name('topology', 'Node'),
        related_name='target_link_set',
        on_delete=models.CASCADE,
    )
    cost = models.FloatField()
    cost_text = models.CharField(max_length=24, blank=True)
    STATUS = Choices('up', 'down')
    status = StatusField()
    properties = JSONField(
        default=dict,
        blank=True,
        load_kwargs={'object_pairs_hook': OrderedDict},
        dump_kwargs={'indent': 4, 'cls': JSONEncoder},
    )
    user_properties = JSONField(
        verbose_name=_('user defined properties'),
        help_text=_('If you need to add additional data to this link use this field'),
        default=dict,
        blank=True,
        load_kwargs={'object_pairs_hook': OrderedDict},
        dump_kwargs={'indent': 4, 'cls': JSONEncoder},
    )
    status_changed = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._initial_status = self.status

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        if self.status != self._initial_status:
            self.send_status_changed_signal()
            self.status_changed = now()
        self._initial_status = self.status

    def __str__(self):
        return '{0} - {1}'.format(self.source.get_name(), self.target.get_name())

    def full_clean(self, *args, **kwargs):
        self.organization = self.topology.organization
        return super().full_clean(*args, **kwargs)

    def clean(self):
        if self.source == self.target or self.source_id == self.target_id:
            raise ValidationError(_('source and target must not be the same'))
        if self.properties is None:
            self.properties = {}
        # these properties are auto-generated and should not be saved
        for attr in ['status', 'status_changed', 'modified', 'created']:
            if attr in self.properties:
                del self.properties[attr]

    def json(self, dict=False, original=False, **kwargs):
        """
        Returns a NetJSON NetworkGraph Link object.

        If ``original`` is passed, the data will be returned
        as it has been collected from the network (used when
        doing the comparison).
        """
        netjson = OrderedDict(
            (
                ('source', self.source.netjson_id),
                ('target', self.target.netjson_id),
                ('cost', self.cost),
                ('cost_text', self.cost_text or ''),
            )
        )
        # properties contain status by default
        properties = OrderedDict()
        if self.properties:
            properties.update(self.properties)
        if not original:
            properties.update(self.user_properties)
            properties['status'] = self.status
            properties['status_changed'] = self.status_changed
            properties['created'] = self.created
            properties['modified'] = self.modified
        netjson['properties'] = properties
        if dict:
            return netjson
        return json.dumps(netjson, cls=JSONEncoder, **kwargs)

    def send_status_changed_signal(self):
        link_status_changed.send(sender=self.__class__, link=self)

    @classmethod
    def get_from_nodes(cls, source, target, topology):
        """
        Find link between source and target,
        (or vice versa, order is irrelevant).
        Source and target nodes must already exist.
        :param source: string
        :param target: string
        :param topology: Topology instance
        :returns: Link object or None
        """
        source = '"{}"'.format(source)
        target = '"{}"'.format(target)
        q = Q(
            source__addresses__contains=source, target__addresses__contains=target
        ) | Q(source__addresses__contains=target, target__addresses__contains=source)
        return cls.objects.filter(q).filter(topology=topology).first()

    @classmethod
    def delete_expired_links(cls):
        """
        deletes links that have been down for more than
        ``NETJSONGRAPH_LINK_EXPIRATION`` days
        """
        LINK_EXPIRATION = app_settings.LINK_EXPIRATION
        if LINK_EXPIRATION not in [False, None]:
            expiration_date = now() - timedelta(days=int(LINK_EXPIRATION))
            expired_links = cls.objects.filter(
                status='down', modified__lt=expiration_date
            )
            expired_links_length = len(expired_links)
            if expired_links_length:
                print_info('Deleting {0} expired links'.format(expired_links_length))
                for link in expired_links:
                    link.delete()

    @classmethod
    def get_queryset(cls, qs):
        """ admin list queryset """
        return qs.select_related('organization', 'topology', 'source', 'target')



import sys

from django.conf.urls import url
from django.core.exceptions import ValidationError
from django.dispatch import Signal
from django.http import Http404
from django.shortcuts import get_object_or_404 as get_obj_or_404

link_status_changed = Signal(providing_args=["link"])


def print_info(message):  # pragma no cover
    """
    print info message if calling from management command ``update_all``
    """
    if 'update_topology' in sys.argv:
        print('{0}\n'.format(message))


def get_object_or_404(model, pk, **kwargs):
    """
    retrieves topology with specified arguments or raises 404
    """
    kwargs.update({'pk': pk, 'published': True})
    try:
        return get_obj_or_404(model, **kwargs)
    except ValidationError:
        raise Http404()


def get_api_urls(views_module):
    """
    used by third party apps to reduce boilerplate
    """
    urls = [
        url(r'^topology/$', views_module.network_collection, name='network_collection'),
        url(
            r'^topology/(?P<pk>[^/]+)/$',
            views_module.network_graph,
            name='network_graph',
        ),
        url(
            r'^topology/(?P<pk>[^/]+)/history/$',
            views_module.network_graph_history,
            name='network_graph_history',
        ),
        url(
            r'^receive/(?P<pk>[^/\?]+)/$',
            views_module.receive_topology,
            name='receive_topology_deprecated',
        ),
        url(
            r'^topology/(?P<pk>[^/\?]+)/receive/$',
            views_module.receive_topology,
            name='receive_topology',
        ),
    ]
    return urls


def get_visualizer_urls(views_module):
    """
    used by third party apps to reduce boilerplate
    """
    urls = [
        url(r'^$', views_module.topology_list, name='topology_list'),
        url(
            r'^topology/(?P<pk>[^/]+)/$',
            views_module.topology_detail,
            name='topology_detail',
        ),
    ]
    return urls



# Generated by Django 2.2.14 on 2020-07-15 21:26

import uuid

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.CONFIG_DEVICE_MODEL),
        ('sample_network_topology', '0002_json_properties'),
    ]

    operations = [
        migrations.CreateModel(
            name='DeviceNode',
            fields=[
                (
                    'id',
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ('is_test', models.BooleanField(default=True)),
                (
                    'device',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.CONFIG_DEVICE_MODEL,
                    ),
                ),
                (
                    'node',
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        to='sample_network_topology.Node',
                    ),
                ),
            ],
            options={'abstract': False, 'unique_together': {('node', 'device')}},
        ),
    ]



from unittest.mock import patch

import swapper
from django.contrib.auth.models import Permission
from django.test import TestCase
from django.urls import reverse

from openwisp_users.tests.utils import TestOrganizationMixin

from .. import settings as app_settings
from .utils import CreateGraphObjectsMixin, UnpublishMixin

Node = swapper.load_model('topology', 'Node')
Topology = swapper.load_model('topology', 'Topology')
OrganizationUser = swapper.load_model('openwisp_users', 'OrganizationUser')


class TestVisualizer(
    UnpublishMixin, CreateGraphObjectsMixin, TestOrganizationMixin, TestCase
):
    topology_model = Topology
    node_model = Node

    def setUp(self):
        org = self._create_org()
        user = self._create_admin(username='visualizer', email='visualizer@email.com')
        self._create_org_user(user=user, organization=org)
        self.client.force_login(user)
        t = self._create_topology(organization=org)
        self._create_node(
            label='node1', addresses=['192.168.0.1'], topology=t, organization=org
        )
        self._create_node(
            label='node2', addresses=['192.168.0.2'], topology=t, organization=org
        )

    def test_list(self):
        response = self.client.get(reverse('topology_list'))
        self.assertContains(response, 'TestNetwork')

    def test_detail(self):
        t = self.topology_model.objects.first()
        response = self.client.get(reverse('topology_detail', args=[t.pk]))
        self.assertContains(response, t.pk)
        # ensure switcher is present
        self.assertContains(response, 'njg-switcher')
        self.assertContains(response, 'njg-datepicker')

    def test_list_unpublished(self):
        self._unpublish()
        response = self.client.get(reverse('topology_list'))
        self.assertNotContains(response, 'TestNetwork')

    def test_detail_unpublished(self):
        self._unpublish()
        t = self.topology_model.objects.first()
        response = self.client.get(reverse('topology_detail', args=[t.pk]))
        self.assertEqual(response.status_code, 404)

    def test_detail_uuid_exception(self):
        """
        see https://github.com/netjson/django-netjsongraph/issues/4
        """
        t = self.topology_model.objects.first()
        response = self.client.get(
            reverse('topology_detail', args=['{0}-wrong'.format(t.pk)])
        )
        self.assertEqual(response.status_code, 404)

    def _test_visualizer_with_unauthenticated_user(self, url):
        self.client.logout()
        r = self.client.get(url)
        self.assertContains(
            r, 'Authentication credentials were not provided.', status_code=403
        )

    def _test_visualizer_with_not_a_manager_user(self, user, url, is_detail=False):
        OrganizationUser.objects.filter(user=user).delete()
        perm = Permission.objects.filter(codename__endswith='topology')
        user.user_permissions.add(*perm)
        self.client.force_login(user)
        r = self.client.get(url)
        if is_detail:
            detail = (
                'User is not a manager of the organization to '
                'which the requested resource belongs.'
            )
            self.assertContains(r, detail, status_code=403)
        else:
            self.assertEqual(r.status_code, 200)
            self.assertNotContains(r, 'TestNetwork')
            self.assertNotEqual(Topology.objects.all().count(), 0)

    def _test_visualizer_with_not_permitted_user(self, user, url):
        t = self.topology_model.objects.first()
        self._create_org_user(user=user, organization=t.organization, is_admin=True)
        self.client.force_login(user)
        r = self.client.get(url)
        self.assertContains(
            r, 'You do not have permission to perform this action.', status_code=403
        )

    def test_list_visualizer_with_auth_enabled(self):
        user = self._create_user(username='list-user', email='list@email.com')
        url = reverse('topology_list')

        with self.subTest('test list visualizer with not auth user'):
            self._test_visualizer_with_unauthenticated_user(url)

        with self.subTest('test list visualizer with not permitted user'):
            self._test_visualizer_with_not_permitted_user(user, url)

        with self.subTest('test list visualizer with not a manager user'):
            self._test_visualizer_with_not_a_manager_user(user, url)

    def test_detail_visualizer_with_auth_enabled(self):
        user = self._create_user(username='detail-user', email='detail@email.com')
        t = self.topology_model.objects.first()
        url = reverse('topology_detail', args=[t.pk])

        with self.subTest('test detail visualizer with not auth user'):
            self._test_visualizer_with_unauthenticated_user(url)

        with self.subTest('test detail visualizer with not permitted user'):
            self._test_visualizer_with_not_permitted_user(user, url)

        with self.subTest('test detail visualizer with not a manager user'):
            self._test_visualizer_with_not_a_manager_user(user, url, is_detail=True)

    def _successful_visualizer_tests(self):
        with self.subTest('test superuser on list visualizer'):
            r = self.client.get(reverse('topology_list'))
            self.assertContains(r, 'TestNetwork')

        with self.subTest('test superuser on detail visualizer'):
            t = self.topology_model.objects.first()
            r = self.client.get(reverse('topology_detail', args=[t.pk]))
            self.assertContains(r, t.pk)

    def test_visualizer_with_auth_enabled_superuser(self):
        user = self._create_admin(username='super-visualizer', email='s@email.com')
        self.client.force_login(user)
        self._successful_visualizer_tests()

    @patch.object(app_settings, 'TOPOLOGY_API_AUTH_REQUIRED', return_value=False)
    def test_visualizer_with_auth_disabled(self, mocked):
        self._successful_visualizer_tests()

    @patch.object(app_settings, 'TOPOLOGY_API_AUTH_REQUIRED', return_value=False)
    def test_visualizer_with_auth_disabled_superuser(self, mocked):
        user = self._create_admin(username='super-visualizer', email='s@email.com')
        self.client.force_login(user)
        self._successful_visualizer_tests()



from openwisp_network_topology.tests.test_admin import TestAdmin as BaseTestAdmin
from openwisp_network_topology.tests.test_admin import (
    TestMultitenantAdmin as BaseTestMultitenantAdmin,
)
from openwisp_network_topology.tests.test_api import TestApi as BaseTestApi
from openwisp_network_topology.tests.test_link import TestLink as BaseTestLink
from openwisp_network_topology.tests.test_node import TestNode as BaseTestNode
from openwisp_network_topology.tests.test_snapshot import (
    TestSnapshot as BaseTestSnapshot,
)
from openwisp_network_topology.tests.test_topology import (
    TestTopology as BaseTestTopology,
)
from openwisp_network_topology.tests.test_upgrader_script import (
    TestUpgradeFromDjangoNetjsongraph as BaseTestUpgradeFromDjangoNetjsongraph,
)
from openwisp_network_topology.tests.test_utils import TestUtils as BaseTestUtils
from openwisp_network_topology.tests.test_visualizer import (
    TestVisualizer as BaseTestVisualizer,
)
from openwisp_users.tests.test_admin import TestBasicUsersIntegration


class TestAdmin(BaseTestAdmin):
    module = 'openwisp2.sample_network_topology'
    app_label = 'sample_network_topology'


class TestMultitenantAdmin(BaseTestMultitenantAdmin):
    app_label = 'sample_network_topology'


class TestApi(BaseTestApi):
    pass


class TestLink(BaseTestLink):
    pass


class TestNode(BaseTestNode):
    pass


class TestTopology(BaseTestTopology):
    pass


class TestUtils(BaseTestUtils):
    pass


class TestVisualizer(BaseTestVisualizer):
    pass


class TestSnapshot(BaseTestSnapshot):
    pass


class TestUpgradeFromDjangoNetjsongraph(BaseTestUpgradeFromDjangoNetjsongraph):
    pass


class TestUsersIntegration(TestBasicUsersIntegration):
    pass


del BaseTestAdmin
del BaseTestMultitenantAdmin
del BaseTestApi
del BaseTestLink
del BaseTestNode
del BaseTestSnapshot
del BaseTestTopology
del BaseTestUtils
del BaseTestVisualizer
del BaseTestUpgradeFromDjangoNetjsongraph
del TestBasicUsersIntegration



import swapper

from .base.link import AbstractLink
from .base.node import AbstractNode
from .base.snapshot import AbstractSnapshot
from .base.topology import AbstractTopology


class Link(AbstractLink):
    class Meta(AbstractLink.Meta):
        abstract = False
        swappable = swapper.swappable_setting('topology', 'Link')


class Node(AbstractNode):
    class Meta(AbstractNode.Meta):
        abstract = False
        swappable = swapper.swappable_setting('topology', 'Node')


class Snapshot(AbstractSnapshot):
    class Meta(AbstractSnapshot.Meta):
        abstract = False
        swappable = swapper.swappable_setting('topology', 'Snapshot')


class Topology(AbstractTopology):
    class Meta(AbstractTopology.Meta):
        abstract = False
        swappable = swapper.swappable_setting('topology', 'Topology')



import swapper
from django.db import models
from django.utils.translation import ugettext_lazy as _

from openwisp_users.mixins import OrgMixin
from openwisp_utils.base import TimeStampedEditableModel


class AbstractSnapshot(OrgMixin, TimeStampedEditableModel):
    """
    NetJSON NetworkGraph Snapshot implementation
    """

    topology = models.ForeignKey(
        swapper.get_model_name('topology', 'Topology'), on_delete=models.CASCADE
    )
    data = models.TextField(blank=False)
    date = models.DateField(auto_now=True)

    class Meta:
        verbose_name_plural = _('snapshots')
        abstract = True

    def __str__(self):
        return "{0}: {1}".format(self.topology.label, self.date)



from django.db import migrations

from . import fix_link_properties


class Migration(migrations.Migration):
    dependencies = [
        ('topology', '0010_properties_json'),
    ]

    operations = [
        migrations.RunPython(
            fix_link_properties, reverse_code=migrations.RunPython.noop
        )
    ]



import os

from celery import Celery

os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'openwisp2.settings')

app = Celery('openwisp2')
app.config_from_object('django.conf:settings', namespace='CELERY')
app.autodiscover_tasks()



import swapper
from django.core.management.base import BaseCommand


class BaseCreateDeviceNodeCommand(BaseCommand):
    help = 'Create initial DeviceNode objects'

    def handle(self, *args, **kwargs):
        Node = swapper.load_model('topology', 'Node')
        DeviceNode = swapper.load_model('topology_device', 'DeviceNode')
        queryset = Node.objects.select_related('topology').filter(
            topology__parser='netdiff.OpenvpnParser'
        )
        for node in queryset.iterator():
            DeviceNode.auto_create(node)



from django.conf import settings
from django.core.exceptions import ValidationError
from rest_framework.exceptions import ParseError
from rest_framework.parsers import JSONParser


class TextParser(JSONParser):
    """
    Dummy TextParser accepting any text (used in ReceiveTopologyView)
    """

    media_type = 'text/plain'

    def parse(self, stream, media_type=None, parser_context=None):
        parser_context = parser_context or {}
        encoding = parser_context.get('encoding', settings.DEFAULT_CHARSET)

        try:
            return stream.read().decode(encoding)
        except ValidationError as e:  # pragma: nocover
            raise ParseError('text/plain parse error - %s' % str(e))



from django.db import models

from openwisp_network_topology.base.link import AbstractLink
from openwisp_network_topology.base.node import AbstractNode
from openwisp_network_topology.base.snapshot import AbstractSnapshot
from openwisp_network_topology.base.topology import AbstractTopology


class DetailsModel(models.Model):
    details = models.CharField(max_length=64, blank=True, null=True)

    class Meta:
        abstract = True


class Link(DetailsModel, AbstractLink):
    class Meta(AbstractLink.Meta):
        abstract = False


class Node(DetailsModel, AbstractNode):
    class Meta(AbstractNode.Meta):
        abstract = False


class Snapshot(DetailsModel, AbstractSnapshot):
    class Meta(AbstractSnapshot.Meta):
        abstract = False


class Topology(DetailsModel, AbstractTopology):
    class Meta(AbstractTopology.Meta):
        abstract = False
